# copyright 2004-2021 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of rql.
#
# rql is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# rql is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with rql. If not, see <http://www.gnu.org/licenses/>.

import unittest

from logilab.common.testlib import TestCase, mock_object as mock

from rql import RQLHelper, TypeResolverException

FINAL_ETYPES = ("String", "Boolean", "Int", "Float", "Date", "Datetime")


class ERSchema:
    def __hash__(self):
        return hash(self.type)

    def __str__(self):
        return self.type


class RelationSchema(ERSchema):
    def __init__(self, assoc_types, symmetric=False, card=None):
        self.assoc_types = assoc_types
        self.subj_types = [e_type[0] for e_type in assoc_types]
        d = {}
        for e_type, dest_types in assoc_types:
            for e_type in dest_types:
                d[e_type] = 1
        self.obj_types = list(d.keys())
        self.symmetric = symmetric
        self.inlined = False
        if card is None:
            if self.final:
                card = "?1"
            else:
                card = "**"
        self.card = card
        self.rdefs = {}
        for subjtype, dest_types in self.assoc_types:
            for objtype in dest_types:
                self.rdefs[(subjtype, objtype)] = mock(
                    subject=subjtype, object=objtype, cardinality=self.card
                )

    def associations(self):
        return self.assoc_types

    def subjects(self, etype=None):
        return self.subj_types

    def objects(self, etype=None):
        return self.obj_types

    @property
    def final(self):
        return self.obj_types[0] in FINAL_ETYPES


class EntitySchema(ERSchema):
    def __init__(self, type, specialized_by=None):
        self.type = type
        self._specialized_by = specialized_by or ()
        self.final = self.type in FINAL_ETYPES

    def specialized_by(self):
        return self._specialized_by


class DummySchema:
    def __init__(self):
        self._types = {}
        for etype in [
            "String",
            "Boolean",
            "Int",
            "Float",
            "Date",
            "Datetime",
            "Eetype",
            "Person",
            "Company",
            "Address",
            "Student",
        ]:
            self._types[etype] = EntitySchema(etype)
        self._types["Person"]._specialized_by = [self._types["Student"]]
        self._relations = {
            "eid": RelationSchema(
                (
                    ("Person", ("Int",)),
                    ("Student", ("Int",)),
                    ("Company", ("Int",)),
                    ("Address", ("Int",)),
                    ("Eetype", ("Int",)),
                )
            ),
            "creation_date": RelationSchema(
                (
                    ("Person", ("Datetime",)),
                    ("Student", ("Datetime",)),
                    ("Company", ("Datetime",)),
                    ("Address", ("Datetime",)),
                    ("Eetype", ("Datetime",)),
                )
            ),
            "name": RelationSchema(
                (
                    ("Person", ("String",)),
                    ("Student", ("String",)),
                    ("Company", ("String",)),
                )
            ),
            "firstname": RelationSchema(
                (
                    ("Person", ("String",)),
                    ("Student", ("String",)),
                )
            ),
            "work_for": RelationSchema(
                (
                    ("Person", ("Company",)),
                    ("Student", ("Company",)),
                ),
                card="?*",
            ),
            "is": RelationSchema(
                (
                    ("Person", ("Eetype",)),
                    ("Student", ("Eetype",)),
                    ("Company", ("Eetype",)),
                    ("Address", ("Eetype",)),
                    ("Eetype", ("Eetype",)),
                )
            ),
            "is_instance_of": RelationSchema(
                (
                    ("Person", ("Eetype",)),
                    ("Student", ("Eetype",)),
                    ("Company", ("Eetype",)),
                    ("Address", ("Eetype",)),
                    ("Eetype", ("Eetype",)),
                )
            ),
            "connait": RelationSchema(
                (
                    ("Person", ("Person",)),
                    ("Student", ("Person",)),
                    ("Student", ("Student",)),
                    ("Person", ("Student",)),
                ),
                symmetric=True,
            ),
            "located": RelationSchema(
                (
                    ("Person", ("Address",)),
                    ("Student", ("Address",)),
                    ("Company", ("Address",)),
                )
            ),
            "owned_by": RelationSchema(
                (
                    ("Person", ("Person",)),
                    ("Student", ("Person",)),
                    ("Company", ("Person",)),
                    ("Eetype", ("Person",)),
                )
            ),
            "identity": RelationSchema(
                (
                    ("Person", ("Person",)),
                    ("Student", ("Student",)),
                    ("Company", ("Company",)),
                    ("Address", ("Address",)),
                    ("Eetype", ("Eetype",)),
                )
            ),
            "number": RelationSchema(
                (
                    ("Person", ("Int",)),
                    ("Student", ("Int",)),
                    ("Company", ("Float",)),
                )
            ),
        }
        for rtype, rschema in self._relations.items():
            rschema.type = rtype

    def entities(self):
        return self._types.values()

    def relations(self):
        return self._relations.keys()

    def has_entity(self, e_type):
        return e_type in self._types

    def has_relation(self, r_type):
        return r_type in self._relations

    def __contains__(self, ertype):
        return self.has_entity(ertype) or self.has_relation(ertype)

    def relation_schema_for(self, r_type: str):
        return self._relations[r_type]

    def rschema(self, r_type):
        return self.relation_schema_for(r_type)

    def entity_schema_for(self, e_type: str):
        return self._types[e_type]

    def eschema(self, e_type: str):
        return self.entity_schema_for(e_type)


UNRESOLVABLE_QUERIES = (
    "Person X WHERE Y work_for X",
    "Person X WHERE X work_for Y, Y is Address",
    'Insert Company X : X name "toto", X work_for Y WHERE Y name "logilab"',
)

DEBUG = 0
ALL_SOLS = [
    {"X": "Address"},
    {"X": "Company"},
    {"X": "Eetype"},
    {"X": "Person"},
    {"X": "Student"},
]


class AnalyzerClassTest(TestCase):
    """check wrong queries arre correctly detected"""

    eids = {10: "Eetype"}

    def _type_from_eid(self, eid):
        return self.eids.get(eid, "Person")

    def setUp(self):
        self.helper = RQLHelper(DummySchema(), {"eid": self._type_from_eid})

    def test_raise(self):
        for rql in UNRESOLVABLE_QUERIES:
            if DEBUG:
                print(rql)
            node = self.helper.parse(rql)
            self.assertRaises(
                TypeResolverException, self.helper.compute_solutions, node, debug=DEBUG
            )

    def test_base_1(self):
        node = self.helper.parse("Any X")
        self.helper.compute_solutions(node, debug=DEBUG)
        sols = node.children[0].solutions
        self.assertCountEqual(
            sols,
            [
                {"X": "Address"},
                {"X": "Company"},
                {"X": "Eetype"},
                {"X": "Person"},
                {"X": "Student"},
            ],
        )

    def test_base_2(self):
        node = self.helper.parse("Person X")
        # check constant type of the is relation inserted
        self.assertEqual(node.children[0].where.children[1].children[0].type, "etype")
        self.helper.compute_solutions(node, debug=DEBUG)
        sols = node.children[0].solutions
        self.assertEqual(sols, [{"X": "Person"}])

    def test_base_3(self):
        node = self.helper.parse("Any X WHERE X eid 1")
        self.helper.compute_solutions(node, debug=DEBUG)
        sols = node.children[0].solutions
        self.assertEqual(sols, [{"X": "Person"}])
        self.helper.simplify(node)
        self.helper.compute_solutions(node, debug=DEBUG)
        sols = node.children[0].solutions
        self.assertEqual(sols, [{}])

    def test_base_guess_1(self):
        node = self.helper.parse("Person X WHERE X work_for Y")
        self.helper.compute_solutions(node, debug=DEBUG)
        sols = node.children[0].solutions
        self.assertCountEqual(sols, [{"X": "Person", "Y": "Company"}])

    def test_base_guess_2(self):
        node = self.helper.parse('Any X WHERE X name "Logilab"')
        self.helper.compute_solutions(node, debug=DEBUG)
        sols = node.children[0].solutions
        self.assertCountEqual(
            sols, [{"X": "Company"}, {"X": "Person"}, {"X": "Student"}]
        )

    def test_non_regr_no_final_type(self):
        """https://www.logilab.net/elo/ticket/9042"""
        node = self.helper.parse("Any X WHERE X creation_date > ((2009 - 4) - 16)")
        self.helper.compute_solutions(node, debug=DEBUG)
        sols = node.children[0].solutions
        self.assertCountEqual(
            sols,
            [
                {"X": "Address"},
                {"X": "Company"},
                {"X": "Eetype"},
                {"X": "Person"},
                {"X": "Student"},
            ],
        )

    def test_is_instance_of_1(self):
        node = self.helper.parse("Any X WHERE X is_instance_of Person")
        # check constant type of the is relation inserted
        self.assertEqual(node.children[0].where.children[1].children[0].type, "etype")
        self.helper.compute_solutions(node, debug=DEBUG)
        sols = node.children[0].solutions
        self.assertCountEqual(sols, [{"X": "Person"}, {"X": "Student"}])

    def test_is_instance_of_2(self):
        node = self.helper.parse("Any X WHERE X is_instance_of Student")
        # check constant type of the is relation inserted
        self.assertEqual(node.children[0].where.children[1].children[0].type, "etype")
        self.helper.compute_solutions(node, debug=DEBUG)
        sols = node.children[0].solutions
        self.assertEqual(sols, [{"X": "Student"}])

    def test_is_query(self):
        node = self.helper.parse('Any T WHERE X name "logilab", X is T')
        self.helper.compute_solutions(node, debug=DEBUG)
        sols = node.children[0].solutions
        self.assertCountEqual(
            sols,
            [
                {"X": "Company", "T": "Eetype"},
                {"X": "Person", "T": "Eetype"},
                {"X": "Student", "T": "Eetype"},
            ],
        )

    def test_is_query_const(self):
        node = self.helper.parse("Any X WHERE X is T, T eid 10")
        self.helper.compute_solutions(node, debug=DEBUG)
        sols = node.children[0].solutions
        self.assertCountEqual(
            sols,
            [
                {"X": "Address", "T": "Eetype"},
                {"X": "Company", "T": "Eetype"},
                {"X": "Eetype", "T": "Eetype"},
                {"X": "Person", "T": "Eetype"},
                {"X": "Student", "T": "Eetype"},
            ],
        )

    def test_not(self):
        node = self.helper.parse("Any X WHERE NOT X is Person")
        self.helper.compute_solutions(node, debug=DEBUG)
        sols = node.children[0].solutions
        expected = ALL_SOLS[:]
        expected.remove({"X": "Person"})
        self.assertCountEqual(sols, expected)

    def test_not_identity(self):
        node = self.helper.parse(
            "Any X WHERE X located A, P is Person, NOT X identity P"
        )
        self.helper.compute_solutions(node, debug=DEBUG)
        sols = node.children[0].solutions
        self.assertCountEqual(
            sols,
            [
                {"X": "Company", "A": "Address", "P": "Person"},
                {"X": "Person", "A": "Address", "P": "Person"},
                {"X": "Student", "A": "Address", "P": "Person"},
            ],
        )

    def test_uid_func_mapping(self):
        h = self.helper

        def type_from_uid(name):
            self.assertEqual(name, "Logilab")
            return "Company"

        uid_func_mapping = {"name": type_from_uid}
        # constant as rhs of the uid relation
        node = h.parse('Any X WHERE X name "Logilab"')
        h.compute_solutions(node, uid_func_mapping, debug=DEBUG)
        sols = node.children[0].solutions
        self.assertCountEqual(sols, [{"X": "Company"}])
        # variable as rhs of the uid relation
        node = h.parse("Any N WHERE X name N")
        h.compute_solutions(node, uid_func_mapping, debug=DEBUG)
        sols = node.children[0].solutions
        self.assertCountEqual(
            sols,
            [
                {"X": "Company", "N": "String"},
                {"X": "Person", "N": "String"},
                {"X": "Student", "N": "String"},
            ],
        )
        # substitute as rhs of the uid relation
        node = h.parse("Any X WHERE X name %(company)s")
        h.compute_solutions(node, uid_func_mapping, {"company": "Logilab"}, debug=DEBUG)
        sols = node.children[0].solutions
        self.assertCountEqual(sols, [{"X": "Company"}])

    def test_non_regr_subjobj1(self):
        h = self.helper

        def type_from_uid(name):
            self.assertEqual(name, "Societe")
            return "Eetype"

        uid_func_mapping = {"name": type_from_uid}
        # constant as rhs of the uid relation
        node = h.parse('Any X WHERE X name "Societe", X is ISOBJ, ISSIBJ is X')
        h.compute_solutions(node, uid_func_mapping, debug=DEBUG)
        sols = node.children[0].solutions
        self.assertCountEqual(
            sols,
            [
                {"X": "Eetype", "ISOBJ": "Eetype", "ISSIBJ": "Address"},
                {"X": "Eetype", "ISOBJ": "Eetype", "ISSIBJ": "Company"},
                {"X": "Eetype", "ISOBJ": "Eetype", "ISSIBJ": "Eetype"},
                {"X": "Eetype", "ISOBJ": "Eetype", "ISSIBJ": "Person"},
                {"X": "Eetype", "ISOBJ": "Eetype", "ISSIBJ": "Student"},
            ],
        )

    def test_non_regr_subjobj2(self):
        h = self.helper

        def type_from_uid(name):
            self.assertEqual(name, "Societe")
            return "Eetype"

        uid_func_mapping = {"name": type_from_uid}
        node = h.parse(
            'Any X WHERE X name "Societe", X is ISOBJ, ISSUBJ is X, '
            "X is_instance_of ISIOOBJ, ISIOSUBJ is_instance_of X"
        )
        h.compute_solutions(node, uid_func_mapping, debug=DEBUG)
        select = node.children[0]
        sols = select.solutions
        self.assertEqual(len(sols), 25)

        def var_sols(var):
            s = set()
            for sol in sols:
                s.add(sol.get(var))
            return s

        self.assertEqual(var_sols("X"), set(("Eetype",)))
        self.assertEqual(
            var_sols("X"), select.defined_vars["X"].stinfo["possibletypes"]
        )
        self.assertEqual(
            var_sols("ISSUBJ"),
            set(("Address", "Company", "Eetype", "Person", "Student")),
        )
        self.assertEqual(
            var_sols("ISSUBJ"),
            select.defined_vars["ISSUBJ"].stinfo["possibletypes"],
        )
        self.assertEqual(var_sols("ISOBJ"), set(("Eetype",)))
        self.assertEqual(
            var_sols("ISOBJ"),
            select.defined_vars["ISOBJ"].stinfo["possibletypes"],
        )
        self.assertEqual(
            var_sols("ISIOSUBJ"),
            set(("Address", "Company", "Eetype", "Person", "Student")),
        )
        self.assertEqual(
            var_sols("ISIOSUBJ"),
            select.defined_vars["ISIOSUBJ"].stinfo["possibletypes"],
        )
        self.assertEqual(var_sols("ISIOOBJ"), set(("Eetype",)))
        self.assertEqual(
            var_sols("ISIOOBJ"),
            select.defined_vars["ISIOOBJ"].stinfo["possibletypes"],
        )

    def test_unusableuid_func_mapping(self):
        h = self.helper

        def type_from_uid(name):
            self.assertEqual(name, "Logilab")
            return "Company"

        uid_func_mapping = {"name": type_from_uid}
        node = h.parse("Any X WHERE NOT X name %(company)s")
        h.compute_solutions(node, uid_func_mapping, {"company": "Logilab"}, debug=DEBUG)
        sols = node.children[0].solutions
        self.assertCountEqual(sols, ALL_SOLS)
        node = h.parse("Any X WHERE X name > %(company)s")
        h.compute_solutions(node, uid_func_mapping, {"company": "Logilab"}, debug=DEBUG)
        sols = node.children[0].solutions
        self.assertCountEqual(sols, ALL_SOLS)

    def test_base_guess_3(self):
        node = self.helper.parse("Any Z GROUPBY Z WHERE X name Z")
        self.helper.compute_solutions(node, debug=DEBUG)
        sols = node.children[0].solutions
        self.assertCountEqual(
            sols,
            [
                {"X": "Company", "Z": "String"},
                {"X": "Person", "Z": "String"},
                {"X": "Student", "Z": "String"},
            ],
        )

    def test_var_name(self):
        node = self.helper.parse("Any E1 GROUPBY E1 WHERE E2 is Person, E2 name E1")
        self.helper.compute_solutions(node, debug=DEBUG)
        sols = node.children[0].solutions
        self.assertCountEqual(sols, [{"E2": "Person", "E1": "String"}])

    def test_relation_eid(self):
        node = self.helper.parse("Any E2 WHERE E2 work_for E1, E2 eid 2")
        self.helper.compute_solutions(node, debug=DEBUG)
        sols = node.children[0].solutions
        self.assertCountEqual(sols, [{"E1": "Company", "E2": "Person"}])
        self.helper.simplify(node)
        self.helper.compute_solutions(node, debug=DEBUG)
        sols = node.children[0].solutions
        self.assertCountEqual(sols, [{"E1": "Company"}])

        node = self.helper.parse("Any E1 WHERE E2 work_for E1, E2 eid 2")
        self.helper.compute_solutions(node, debug=DEBUG)
        sols = node.children[0].solutions
        self.assertCountEqual(sols, [{"E1": "Company", "E2": "Person"}])
        self.helper.simplify(node)
        self.helper.compute_solutions(node, debug=DEBUG)
        sols = node.children[0].solutions
        self.assertCountEqual(sols, [{"E1": "Company"}])

    def test_not_symmetric_relation_eid(self):
        node = self.helper.parse("Any P WHERE X eid 0, NOT X connait P")
        self.helper.compute_solutions(node, debug=DEBUG)
        sols = node.children[0].solutions
        self.assertCountEqual(
            sols, [{"P": "Person", "X": "Person"}, {"P": "Student", "X": "Person"}]
        )
        self.helper.simplify(node)
        self.helper.compute_solutions(node, debug=DEBUG)
        sols = node.children[0].solutions
        self.assertCountEqual(sols, [{"P": "Person"}, {"P": "Student"}])

    def test_union(self):
        node = self.helper.parse(
            "(Any P WHERE X eid 0, X is Person, NOT X connait P) "
            "UNION (Any E1 WHERE E2 work_for E1, E2 eid 2)"
        )
        self.helper.compute_solutions(node, debug=DEBUG)
        sols = node.children[0].solutions
        self.assertCountEqual(
            sols,
            [{"P": "Person", "X": "Person"}, {"P": "Student", "X": "Person"}],
        )
        sols = node.children[1].solutions
        self.assertCountEqual(sols, [{"E1": "Company", "E2": "Person"}])
        self.helper.simplify(node)
        self.helper.compute_solutions(node, debug=DEBUG)
        sols = node.children[0].solutions
        self.assertCountEqual(
            sols,
            [{"P": "Person"}, {"P": "Student"}],
        )
        sols = node.children[1].solutions
        self.assertCountEqual(sols, [{"E1": "Company"}])

    def test_exists(self):
        node = self.helper.parse(
            "Any X WHERE X firstname 'lulu',"
            "EXISTS (X owned_by U, U name 'lulufanclub' OR U name 'managers');"
        )
        self.helper.compute_solutions(node, debug=DEBUG)
        sols = node.children[0].solutions
        self.assertCountEqual(
            sols, [{"X": "Person", "U": "Person"}, {"X": "Student", "U": "Person"}]
        )

    def test_subqueries_base(self):
        node = self.helper.parse(
            "Any L, Y, F WHERE Y located L "
            "WITH Y,F BEING ((Any X,F WHERE X is Person, X firstname F) "
            "UNION (Any X,F WHERE X is Company, X name F))"
        )
        self.helper.compute_solutions(node, debug=DEBUG)
        self.assertEqual(
            node.children[0].with_[0].query.children[0].solutions,
            [{"X": "Person", "F": "String"}],
        )
        self.assertEqual(
            node.children[0].with_[0].query.children[1].solutions,
            [{"X": "Company", "F": "String"}],
        )
        sols = node.children[0].solutions
        self.assertCountEqual(
            sols,
            [
                {"Y": "Company", "L": "Address", "F": "String"},
                {"Y": "Person", "L": "Address", "F": "String"},
            ],
        )

    def test_subqueries_aggregat(self):
        node = self.helper.parse(
            "Any L, SUM(X)*100/Y GROUPBY L "
            "WHERE X is Person, X located L "
            "WITH Y BEING (Any SUM(X) WHERE X is Person)"
        )
        self.helper.compute_solutions(node, debug=DEBUG)
        self.assertEqual(
            node.children[0].with_[0].query.children[0].solutions, [{"X": "Person"}]
        )
        self.assertEqual(
            node.children[0].solutions, [{"X": "Person", "Y": "Person", "L": "Address"}]
        )

    def test_subqueries_outer_filter_type(self):
        # this kind of query may be generated by erudi's facettes box
        node = self.helper.parse(
            "Any L, Y, F WHERE Y located L, Y is Person "
            "WITH Y,F BEING ((Any X,F WHERE X is Person, X firstname F) "
            "UNION (Any X,F WHERE X is Company, X name F))"
        )
        self.helper.compute_solutions(node, debug=DEBUG)
        sols = node.children[0].solutions
        self.assertCountEqual(sols, [{"Y": "Person", "L": "Address", "F": "String"}])
        self.assertEqual(
            node.children[0].with_[0].query.children[0].solutions,
            [{"X": "Person", "F": "String"}],
        )
        # auto-simplification
        self.assertEqual(
            node.as_string(),
            "Any L,Y,F WHERE Y located L, Y is Person WITH Y,F "
            "BEING (Any X,F WHERE X is Person, X firstname F)",
        )
        self.assertEqual(
            node.children[0].with_[0].query.children[0].solutions,
            [{"X": "Person", "F": "String"}],
        )

    def test_insert(self):
        node = self.helper.parse(
            'INSERT Person X : X name "toto", X work_for Y WHERE Y name "logilab"'
        )
        self.helper.compute_solutions(node, debug=DEBUG)
        sols = node.solutions
        self.assertCountEqual(sols, [{"X": "Person", "Y": "Company"}])

    def test_delete(self):
        node = self.helper.parse('DELETE Person X WHERE X name "toto", X work_for Y')
        self.helper.compute_solutions(node, debug=DEBUG)
        sols = node.solutions
        self.assertCountEqual(sols, [{"X": "Person", "Y": "Company"}])

    def test_set(self):
        node = self.helper.parse(
            'SET X name "toto", X work_for Y WHERE Y name "logilab"'
        )
        self.helper.compute_solutions(node, debug=DEBUG)
        sols = node.solutions
        self.assertCountEqual(
            sols, [{"X": "Person", "Y": "Company"}, {"X": "Student", "Y": "Company"}]
        )

    def test_set_mathexpr(self):
        node = self.helper.parse("SET S number N/4 WHERE P work_for S, P number N")
        self.helper.compute_solutions(node, debug=DEBUG)
        sols = node.solutions
        self.assertCountEqual(
            sols,
            [
                {"P": "Person", "S": "Company", "N": "Int"},
                {"P": "Student", "S": "Company", "N": "Int"},
            ],
        )

    def test_nonregr_not_u_ownedby_u(self):
        node = self.helper.parse("Any U WHERE NOT U owned_by U")
        self.helper.compute_solutions(node, debug=DEBUG)
        sols = node.children[0].solutions
        self.assertCountEqual(sols, [{"U": "Person"}])

    def test_selection_with_cast(self):
        node = self.helper.parse(
            "Any X WHERE X name CAST(String, E), Y eid E, X owned_by Y"
        )
        self.helper.compute_solutions(node, debug=DEBUG)
        sols = node.children[0].solutions
        self.assertCountEqual(
            sols,
            [
                {"E": "Int", "X": "Company", "Y": "Person"},
                {"E": "Int", "X": "Person", "Y": "Person"},
                {"E": "Int", "X": "Student", "Y": "Person"},
            ],
        )

    def test_set_with_cast(self):
        node = self.helper.parse(
            "SET X name CAST(String, E), X work_for Y WHERE Y eid E"
        )
        self.helper.compute_solutions(node, debug=DEBUG)
        sols = node.solutions
        self.assertCountEqual(
            sols,
            [
                {"X": "Person", "Y": "Company", "E": "Int"},
                {"X": "Student", "Y": "Company", "E": "Int"},
            ],
        )


if __name__ == "__main__":
    unittest.main()



# copyright 2004-2021 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of rql.
#
# rql is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# rql is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with rql. If not, see <http://www.gnu.org/licenses/>.
"""Miscellaneous utilities for RQL."""

import string

from logilab.database import SQL_FUNCTIONS_REGISTRY, FunctionDescr, CAST
from logilab.common.decorators import monkeypatch

from rql._exceptions import BadRQLQuery
from typing import (
    TYPE_CHECKING,
    Set,
    Optional,
    Mapping,
    Any,
    Callable,
    Generator,
    List,
    Union as Union_,
)

if TYPE_CHECKING:
    import rql
    import logilab

    from rql.base import BaseNode

__docformat__: str = "restructuredtext en"


def decompose_b26(index: int, table: str = string.ascii_uppercase) -> str:
    """Return a letter (base-26) decomposition of index."""
    div, mod = divmod(index, 26)

    if div == 0:
        return table[mod]

    return decompose_b26(div - 1) + table[mod]


class rqlvar_maker:
    """Yields consistent RQL variable names.

    :param stop: optional argument to stop iteration after the Nth variable
                 default is None which means 'never stop'
    :param defined: optional dict of already defined vars
    """

    # NOTE: written a an iterator class instead of a simple generator to be
    #       picklable

    def __init__(
        self,
        stop: Optional[int] = None,
        index: int = 0,
        defined: Optional[Mapping[str, Any]] = None,
        aliases: Optional[Mapping[str, Any]] = None,
    ) -> None:
        self.index = index
        self.stop = stop
        self.defined = defined
        self.aliases = aliases

    def __iter__(self):
        return self

    def __next__(self) -> str:
        while self.stop is None or self.index < self.stop:
            var = decompose_b26(self.index)
            self.index += 1
            if self.defined is not None and var in self.defined:
                continue
            if self.aliases is not None and var in self.aliases:
                continue
            return var
        raise StopIteration()

    next = __next__


KEYWORDS: Set[str] = set(
    (
        "INSERT",
        "SET",
        "DELETE",
        "UNION",
        "WITH",
        "BEING",
        "WHERE",
        "AND",
        "OR",
        "NOT",
        "IN",
        "LIKE",
        "ILIKE",
        "EXISTS",
        "DISTINCT",
        "TRUE",
        "FALSE",
        "NULL",
        "TODAY",
        "GROUPBY",
        "HAVING",
        "ORDERBY",
        "ASC",
        "DESC",
        "LIMIT",
        "OFFSET",
    )
)

RQL_FUNCTIONS_REGISTRY: "logilab.database._FunctionRegistry" = (
    SQL_FUNCTIONS_REGISTRY.copy()
)


@monkeypatch(FunctionDescr)
def st_description(  # noqa
    self,
    funcnode: "rql.nodes.Function",
    mainindex: Optional[int],
    tr: Callable[[str], str],
) -> str:
    return "%s(%s)" % (
        tr(self.name),
        ", ".join(
            sorted(
                child.get_description(mainindex, tr)
                for child in iter_funcnode_variables(funcnode)
            )
        ),
    )


@monkeypatch(FunctionDescr)
def st_check_backend(self, backend: Any, funcnode: "rql.nodes.Function") -> None:
    # XXX backend seems to always be None
    if not self.supports(backend):
        raise BadRQLQuery(f"backend {backend} doesn't support function {self.name}")


@monkeypatch(FunctionDescr)
def rql_return_type(self, funcnode: "rql.nodes.Function") -> Optional[str]:
    return self.rtype


@monkeypatch(CAST)  # type: ignore[no-redef] # noqa: F811
def st_description(  # noqa: F811
    self,
    funcnode: "rql.nodes.Function",
    mainindex: Optional[int],
    tr: Callable[[str], str],
) -> str:
    return self.rql_return_type(funcnode)


@monkeypatch(CAST)  # type: ignore[no-redef] # noqa: F811
def rql_return_type(self, funcnode: "rql.nodes.Function") -> str:  # noqa: F811
    # mypy: "BaseNode" has no attribute "value"  [attr-defined]
    # this is black magic to set rql_return_type to logilab.database classes
    # this code assume that CAST only works with nodes that has a value
    # attribute like Constant
    return funcnode.children[0].value  # type: ignore[attr-defined]


def iter_funcnode_variables(funcnode: "rql.nodes.Function") -> Generator:
    # funcnode: rql.nodes.Function
    # term: rql.nodes.VariableRef
    for term in funcnode.children:
        try:
            # term.variable.stinfo: dict
            # mypy: "BaseNode" has no attribute "variable"  [attr-defined]
            # same black magic than in rql_return_type above

            # error: Item "None" of "Optional[Variable]" has no attribute "stinfo"  [union-attr]
            assert getattr(term, "variable") is not None

            yield term.variable.stinfo["attrvar"] or term  # type: ignore[attr-defined]
        except AttributeError:
            yield term


def is_keyword(word: str) -> bool:
    """Return true if the given word is a RQL keyword."""
    return word.upper() in KEYWORDS


def common_parent(
    node1: Optional["BaseNode"], node2: Optional["BaseNode"]
) -> "BaseNode":
    """return the first common parent between node1 and node2

    algorithm :
     1) index node1's parents
     2) climb among node2's parents until we find a common parent
    """
    # index node1's parents
    node1_parents: Set["BaseNode"] = set()
    while node1:
        node1_parents.add(node1)
        node1 = node1.parent
    # climb among node2's parents until we find a common parent
    while node2:
        if node2 in node1_parents:
            return node2
        node2 = node2.parent
    raise Exception(f"Failed to get a common parent between '{node1}' and '{node2}'")


def register_function(funcdef: "logilab.database.FunctionDescr") -> None:
    RQL_FUNCTIONS_REGISTRY.register_function(funcdef)
    SQL_FUNCTIONS_REGISTRY.register_function(funcdef)


def function_description(funcname: str) -> "logilab.database.FunctionDescr":
    # return type is 'logilab.database.' + funcname
    """Return the description (:class:`FunctionDescr`) for a RQL function."""
    return RQL_FUNCTIONS_REGISTRY.get_function(funcname)


def quote(value: str) -> str:
    """Quote a string value."""
    res = ['"']
    for char in value:
        if char == '"':
            res.append("\\")
        res.append(char)
    res.append('"')
    return "".join(res)


def uquote(value: str) -> str:
    # XXX is this still useful?
    """Quote a unicode string value."""
    res: List[str] = ['"']
    for char in value:
        if char == '"':
            res.append("\\")
        res.append(char)
    res.append('"')
    return "".join(res)


class VisitableMixIn:
    def accept(
        self,
        visitor: Union_["rql.stcheck.RQLSTChecker", "rql.stcheck.RQLSTAnnotator"],
        *args: Optional["rql.stcheck.STCheckState"],
        **kwargs: Optional[Any],
    ) -> None:
        visit_id: str = self.__class__.__name__.lower()
        visit_method: Callable = getattr(visitor, f"visit_{visit_id}")
        return visit_method(self, *args, **kwargs)

    def leave(
        self,
        visitor: Union_["rql.stcheck.RQLSTChecker", "rql.stcheck.RQLSTAnnotator"],
        *args: Optional["rql.stcheck.STCheckState"],
        **kwargs: Optional[Any],
    ) -> None:
        visit_id: str = self.__class__.__name__.lower()
        visit_method: Callable = getattr(visitor, f"leave_{visit_id}")
        return visit_method(self, *args, **kwargs)


# should we redefine this as a Protocol?
class RQLVisitorHandler:
    """Handler providing a dummy implementation of all callbacks necessary
    to visit a RQL syntax tree.
    """

    def visit_union(self, union):
        pass

    def visit_insert(self, insert):
        pass

    def visit_delete(self, delete):
        pass

    def visit_set(self, update):
        pass

    def visit_select(self, selection):
        pass

    def visit_sortterm(self, sortterm):
        pass

    def visit_and(self, et):
        pass

    def visit_or(self, ou):
        pass

    def visit_not(self, not_):
        pass

    def visit_relation(self, relation):
        pass

    def visit_comparison(self, comparison):
        pass

    def visit_mathexpression(self, mathexpression):
        pass

    def visit_function(self, function):
        pass

    def visit_variableref(self, variable):
        pass

    def visit_variablerefattributeaccess(self, variable, attribute):
        pass

    def visit_variablerefmethodcall(self, variable, function, args):
        pass

    def visit_constant(self, constant):
        pass



# copyright 2004-2021 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of rql.
#
# rql is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# rql is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with rql. If not, see <http://www.gnu.org/licenses/>.
"""Main parser command.

"""
import argparse
from argparse import RawDescriptionHelpFormatter
from typing import Dict, Any
from rql.parser import Hercule, HerculeScanner
from yapps.runtime import print_error

__docformat__: str = "restructuredtext en"


def main() -> None:
    arg_parser = argparse.ArgumentParser(
        formatter_class=RawDescriptionHelpFormatter,
        description="Parse an RQL string, for example:\n $ python3 -m rql 'Any 1+1;'",
    )
    arg_parser.add_argument("rqlstring", help="the rql string that should be parsed")
    args = arg_parser.parse_args()

    rqlstring = args.rqlstring.strip()

    if not rqlstring.endswith(";"):
        rqlstring += ";"

    parser = Hercule(HerculeScanner(rqlstring))
    e_types: Dict[Any, Any] = {}

    try:
        tree = parser.goal(e_types)
        print("-" * 80)
        print(tree)
        print("-" * 80)
        print(repr(tree))
        print(e_types)
    except SyntaxError as ex:
        print_error(ex, parser._scanner)


main()



"""yapps input grammar for RQL.

:organization: Logilab
:copyright: 2003-2021 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr


Select statement grammar
------------------------

query = <squery> | <union>

union = (<squery>) UNION (<squery>) [UNION (<squery>)]*

squery = Any <selection>
        [GROUPBY <variables>]
        [ORDERBY <sortterms>]
        [LIMIT <nb> OFFSET <nb>]
        [WHERE <restriction>]
        [HAVING <aggregat restriction>]
        [WITH <subquery> [,<subquery]*]

subquery = <variables> BEING (<query>)

variables = <variable> [, <variable>]*


Abbreviations in this code
--------------------------

rules:
* rel -> relation
* decl -> declaration
* expr -> expression
* restr -> restriction
* var -> variable
* func -> function
* const -> constant
* cmp -> comparison

variables:
* R -> syntax tree root
* S -> select node
* P -> parent node

"""

# Begin -- grammar generated by Yapps
from __future__ import print_function
import sys, re
from yapps import runtime


class HerculeScanner(runtime.Scanner):
    patterns = [
        ("'IN'", re.compile("IN")),
        ("'\\)'", re.compile("\\)")),
        ("'\\('", re.compile("\\(")),
        ("','", re.compile(",")),
        ('r"\\)"', re.compile("\\)")),
        ('r"\\("', re.compile("\\(")),
        ('":"', re.compile(":")),
        ("';'", re.compile(";")),
        ("\\s+", re.compile("\\s+")),
        ("/\\*(?:[^*]|\\*(?!/))*\\*/", re.compile("/\\*(?:[^*]|\\*(?!/))*\\*/")),
        ("DELETE", re.compile("(?i)DELETE")),
        ("SET", re.compile("(?i)SET")),
        ("INSERT", re.compile("(?i)INSERT")),
        ("UNION", re.compile("(?i)UNION")),
        ("DISTINCT", re.compile("(?i)DISTINCT")),
        ("WITH", re.compile("(?i)WITH")),
        ("WHERE", re.compile("(?i)WHERE")),
        ("BEING", re.compile("(?i)BEING")),
        ("OR", re.compile("(?i)OR")),
        ("AND", re.compile("(?i)AND")),
        ("NOT", re.compile("(?i)NOT")),
        ("GROUPBY", re.compile("(?i)GROUPBY")),
        ("HAVING", re.compile("(?i)HAVING")),
        ("ORDERBY", re.compile("(?i)ORDERBY")),
        ("SORT_ASC", re.compile("(?i)ASC")),
        ("SORT_DESC", re.compile("(?i)DESC")),
        ("SORT_NULLSFIRST", re.compile("(?i)NULLSFIRST")),
        ("SORT_NULLSLAST", re.compile("(?i)NULLSLAST")),
        ("LIMIT", re.compile("(?i)LIMIT")),
        ("OFFSET", re.compile("(?i)OFFSET")),
        ("DATE", re.compile("(?i)TODAY")),
        ("DATETIME", re.compile("(?i)NOW")),
        ("TRUE", re.compile("(?i)TRUE")),
        ("FALSE", re.compile("(?i)FALSE")),
        ("NULL", re.compile("(?i)NULL")),
        ("EXISTS", re.compile("(?i)EXISTS")),
        ("CMP_OP", re.compile("(?i)<=|<|>=|>|!=|=|~=|LIKE|ILIKE|REGEXP")),
        ("ADD_OP", re.compile("\\+|-|\\||#")),
        ("MUL_OP", re.compile("\\*|/|%|&")),
        ("POW_OP", re.compile("\\^|>>|<<")),
        ("UNARY_OP", re.compile("-|~")),
        ("FUNCTION", re.compile("[A-Za-z_]+\\s*(?=\\()")),
        ("R_TYPE", re.compile("[a-z_][a-z0-9_]*")),
        ("E_TYPE", re.compile("[A-Z][A-Za-z0-9]*[a-z]+[A-Z0-9]*")),
        ("VARIABLE", re.compile("[A-Z][A-Z0-9_]*")),
        ("QMARK", re.compile("\\?")),
        ("FUNCTION_CALL", re.compile("[A-Z][A-Z0-9_]*\\.[a-z_][a-zA-Z0-9_]+")),
        (
            "STRING",
            re.compile("'([^\\'\\\\]|\\\\.)*'|\\\"([^\\\\\\\"\\\\]|\\\\.)*\\\""),
        ),
        ("FLOAT", re.compile("-?\\d+\\.\\d*")),
        ("INT", re.compile("-?\\d+")),
        ("SUBSTITUTE", re.compile("%\\([A-Za-z_0-9]+\\)s")),
    ]

    def __init__(self, str, *args, **kw):
        runtime.Scanner.__init__(
            self,
            None,
            {
                "\\s+": None,
                "/\\*(?:[^*]|\\*(?!/))*\\*/": None,
            },
            str,
            *args,
            **kw
        )


class Hercule(runtime.Parser):
    Context = runtime.Context

    def goal(self, _parent=None):
        _context = self.Context(_parent, self._scanner, "goal", [])
        _token = self._peek(
            "DELETE", "INSERT", "SET", 'r"\\("', "DISTINCT", "E_TYPE", context=_context
        )
        if _token == "DELETE":
            DELETE = self._scan("DELETE", context=_context)
            _delete = self._delete(Delete(), _context)
            self._scan("';'", context=_context)
            return _delete
        elif _token == "INSERT":
            INSERT = self._scan("INSERT", context=_context)
            _insert = self._insert(Insert(), _context)
            self._scan("';'", context=_context)
            return _insert
        elif _token == "SET":
            SET = self._scan("SET", context=_context)
            update = self.update(Set(), _context)
            self._scan("';'", context=_context)
            return update
        else:  # in ['r"\\("', 'DISTINCT', 'E_TYPE']
            union = self.union(Union(), _context)
            self._scan("';'", context=_context)
            return union

    def _delete(self, R, _parent=None):
        _context = self.Context(_parent, self._scanner, "_delete", [R])
        _token = self._peek("E_TYPE", "VARIABLE", context=_context)
        if _token == "VARIABLE":
            decl_rels = self.decl_rels(R, _context)
            where = self.where(R, _context)
            having = self.having(R, _context)
            return R
        else:  # == 'E_TYPE'
            decl_vars = self.decl_vars(R, _context)
            where = self.where(R, _context)
            having = self.having(R, _context)
            return R

    def _insert(self, R, _parent=None):
        _context = self.Context(_parent, self._scanner, "_insert", [R])
        decl_vars = self.decl_vars(R, _context)
        insert_rels = self.insert_rels(R, _context)
        return R

    def insert_rels(self, R, _parent=None):
        _context = self.Context(_parent, self._scanner, "insert_rels", [R])
        _token = self._peek('":"', "';'", context=_context)
        if _token == '":"':
            self._scan('":"', context=_context)
            decl_rels = self.decl_rels(R, _context)
            where = self.where(R, _context)
            having = self.having(R, _context)
            return R
        else:  # == "';'"
            pass

    def update(self, R, _parent=None):
        _context = self.Context(_parent, self._scanner, "update", [R])
        decl_rels = self.decl_rels(R, _context)
        where = self.where(R, _context)
        having = self.having(R, _context)
        return R

    def union(self, R, _parent=None):
        _context = self.Context(_parent, self._scanner, "union", [R])
        _token = self._peek('r"\\("', "DISTINCT", "E_TYPE", context=_context)
        if _token != 'r"\\("':
            select = self.select(Select(), _context)
            R.append(select)
            return R
        else:  # == 'r"\\("'
            self._scan('r"\\("', context=_context)
            select = self.select(Select(), _context)
            self._scan('r"\\)"', context=_context)
            R.append(select)
            while self._peek("UNION", "';'", 'r"\\)"', context=_context) == "UNION":
                UNION = self._scan("UNION", context=_context)
                self._scan('r"\\("', context=_context)
                select = self.select(Select(), _context)
                self._scan('r"\\)"', context=_context)
                R.append(select)
            return R

    def select(self, S, _parent=None):
        _context = self.Context(_parent, self._scanner, "select", [S])
        _token = self._peek("DISTINCT", "E_TYPE", context=_context)
        if _token == "DISTINCT":
            DISTINCT = self._scan("DISTINCT", context=_context)
            select_ = self.select_(S, _context)
            S.distinct = True
            return S
        else:  # == 'E_TYPE'
            select_ = self.select_(S, _context)
            return S

    def select_(self, S, _parent=None):
        _context = self.Context(_parent, self._scanner, "select_", [S])
        E_TYPE = self._scan("E_TYPE", context=_context)
        selection = self.selection(S, _context)
        groupby = self.groupby(S, _context)
        orderby = self.orderby(S, _context)
        limit_offset = self.limit_offset(S, _context)
        where = self.where(S, _context)
        having = self.having(S, _context)
        with_ = self.with_(S, _context)
        S.set_statement_type(E_TYPE)
        return S

    def selection(self, S, _parent=None):
        _context = self.Context(_parent, self._scanner, "selection", [S])
        selection_function = self.selection_function(S, _context)
        while (
            self._peek(
                "','",
                "GROUPBY",
                "ORDERBY",
                "WHERE",
                "LIMIT",
                "OFFSET",
                "HAVING",
                "WITH",
                "';'",
                'r"\\)"',
                context=_context,
            )
            == "','"
        ):
            self._scan("','", context=_context)
            selection_function = self.selection_function(S, _context)

    def selection_function(self, S, _parent=None):
        _context = self.Context(_parent, self._scanner, "selection_function", [S])
        _token = self._peek(
            "FUNCTION_CALL",
            "UNARY_OP",
            'r"\\("',
            "NULL",
            "DATE",
            "DATETIME",
            "TRUE",
            "FALSE",
            "FLOAT",
            "INT",
            "STRING",
            "SUBSTITUTE",
            "VARIABLE",
            "E_TYPE",
            "FUNCTION",
            context=_context,
        )
        if _token != "FUNCTION_CALL":
            expr_add = self.expr_add(S, _context)
            S.append_selected(expr_add)
        else:  # == 'FUNCTION_CALL'
            FUNCTION_CALL = self._scan("FUNCTION_CALL", context=_context)
            _token = self._peek(
                "'\\('",
                "','",
                "GROUPBY",
                "ORDERBY",
                "WHERE",
                "LIMIT",
                "OFFSET",
                "HAVING",
                "WITH",
                "';'",
                'r"\\)"',
                context=_context,
            )
            if _token == "'\\('":
                self._scan("'\\('", context=_context)
                function_args = self.function_args(_context)
                self._scan("'\\)'", context=_context)
                S.append_selected_method_call(FUNCTION_CALL, function_args)
            else:
                S.append_selected_attribute_access(FUNCTION_CALL)

    def function_args(self, _parent=None):
        _context = self.Context(_parent, self._scanner, "function_args", [])
        _token = self._peek(
            "'\\)'",
            "NULL",
            "DATE",
            "DATETIME",
            "TRUE",
            "FALSE",
            "FLOAT",
            "INT",
            "STRING",
            "SUBSTITUTE",
            context=_context,
        )
        if _token != "'\\)'":
            const = self.const(_context)
            args = [const]
            while self._peek("','", "'\\)'", context=_context) == "','":
                self._scan("','", context=_context)
                const = self.const(_context)
                args.append(const)
            return args
        else:  # == "'\\)'"
            return []

    def groupby(self, S, _parent=None):
        _context = self.Context(_parent, self._scanner, "groupby", [S])
        _token = self._peek(
            "GROUPBY",
            "ORDERBY",
            "WHERE",
            "LIMIT",
            "OFFSET",
            "HAVING",
            "WITH",
            "';'",
            'r"\\)"',
            context=_context,
        )
        if _token == "GROUPBY":
            GROUPBY = self._scan("GROUPBY", context=_context)
            nodes = []
            expr_add = self.expr_add(S, _context)
            nodes.append(expr_add)
            while (
                self._peek(
                    "','",
                    "ORDERBY",
                    "WHERE",
                    "LIMIT",
                    "OFFSET",
                    "HAVING",
                    "WITH",
                    "';'",
                    'r"\\)"',
                    context=_context,
                )
                == "','"
            ):
                self._scan("','", context=_context)
                expr_add = self.expr_add(S, _context)
                nodes.append(expr_add)
            S.set_groupby(nodes)
            return True
        else:
            pass

    def having(self, S, _parent=None):
        _context = self.Context(_parent, self._scanner, "having", [S])
        _token = self._peek("HAVING", "WITH", "';'", 'r"\\)"', context=_context)
        if _token == "HAVING":
            HAVING = self._scan("HAVING", context=_context)
            logical_expr = self.logical_expr(S, _context)
            S.set_having([logical_expr])
        else:  # in ['WITH', "';'", 'r"\\)"']
            pass

    def orderby(self, S, _parent=None):
        _context = self.Context(_parent, self._scanner, "orderby", [S])
        _token = self._peek(
            "ORDERBY",
            "WHERE",
            "LIMIT",
            "OFFSET",
            "HAVING",
            "WITH",
            "';'",
            'r"\\)"',
            context=_context,
        )
        if _token == "ORDERBY":
            ORDERBY = self._scan("ORDERBY", context=_context)
            nodes = []
            sort_term = self.sort_term(S, _context)
            nodes.append(sort_term)
            while (
                self._peek(
                    "','",
                    "WHERE",
                    "LIMIT",
                    "OFFSET",
                    "HAVING",
                    "WITH",
                    "';'",
                    'r"\\)"',
                    context=_context,
                )
                == "','"
            ):
                self._scan("','", context=_context)
                sort_term = self.sort_term(S, _context)
                nodes.append(sort_term)
            S.set_orderby(nodes)
            return True
        else:
            pass

    def with_(self, S, _parent=None):
        _context = self.Context(_parent, self._scanner, "with_", [S])
        _token = self._peek("WITH", 'r"\\)"', "';'", context=_context)
        if _token == "WITH":
            WITH = self._scan("WITH", context=_context)
            nodes = []
            subquery = self.subquery(S, _context)
            nodes.append(subquery)
            while self._peek("','", 'r"\\)"', "';'", context=_context) == "','":
                self._scan("','", context=_context)
                subquery = self.subquery(S, _context)
                nodes.append(subquery)
            S.set_with(nodes)
        else:  # in ['r"\\)"', "';'"]
            pass

    def subquery(self, S, _parent=None):
        _context = self.Context(_parent, self._scanner, "subquery", [S])
        variables = self.variables(S, _context)
        node = SubQuery()
        node.set_aliases(variables)
        BEING = self._scan("BEING", context=_context)
        self._scan('r"\\("', context=_context)
        union = self.union(Union(), _context)
        self._scan('r"\\)"', context=_context)
        node.set_query(union)
        return node

    def sort_term(self, S, _parent=None):
        _context = self.Context(_parent, self._scanner, "sort_term", [S])
        expr_add = self.expr_add(S, _context)
        sort_meth = self.sort_meth(_context)
        sort_meth_nulls = self.sort_meth_nulls(_context)
        return SortTerm(expr_add, sort_meth, sort_meth_nulls)

    def sort_meth(self, _parent=None):
        _context = self.Context(_parent, self._scanner, "sort_meth", [])
        _token = self._peek(
            "SORT_DESC",
            "SORT_ASC",
            "SORT_NULLSFIRST",
            "SORT_NULLSLAST",
            "','",
            "WHERE",
            "LIMIT",
            "OFFSET",
            "HAVING",
            "WITH",
            "';'",
            'r"\\)"',
            context=_context,
        )
        if _token == "SORT_DESC":
            SORT_DESC = self._scan("SORT_DESC", context=_context)
            return 0
        elif _token == "SORT_ASC":
            SORT_ASC = self._scan("SORT_ASC", context=_context)
            return 1
        else:
            return 1  # default to SORT_ASC

    def sort_meth_nulls(self, _parent=None):
        _context = self.Context(_parent, self._scanner, "sort_meth_nulls", [])
        _token = self._peek(
            "SORT_NULLSFIRST",
            "SORT_NULLSLAST",
            "','",
            "WHERE",
            "LIMIT",
            "OFFSET",
            "HAVING",
            "WITH",
            "';'",
            'r"\\)"',
            context=_context,
        )
        if _token == "SORT_NULLSFIRST":
            SORT_NULLSFIRST = self._scan("SORT_NULLSFIRST", context=_context)
            return 1
        elif _token == "SORT_NULLSLAST":
            SORT_NULLSLAST = self._scan("SORT_NULLSLAST", context=_context)
            return 2
        else:
            return 0  # default

    def limit_offset(self, R, _parent=None):
        _context = self.Context(_parent, self._scanner, "limit_offset", [R])
        limit = self.limit(R, _context)
        offset = self.offset(R, _context)
        return limit or offset

    def limit(self, R, _parent=None):
        _context = self.Context(_parent, self._scanner, "limit", [R])
        _token = self._peek(
            "LIMIT",
            "OFFSET",
            "WHERE",
            "HAVING",
            "WITH",
            "';'",
            'r"\\)"',
            context=_context,
        )
        if _token == "LIMIT":
            LIMIT = self._scan("LIMIT", context=_context)
            INT = self._scan("INT", context=_context)
            R.set_limit(int(INT))
            return True
        else:  # in ['OFFSET', 'WHERE', 'HAVING', 'WITH', "';'", 'r"\\)"']
            pass

    def offset(self, R, _parent=None):
        _context = self.Context(_parent, self._scanner, "offset", [R])
        _token = self._peek(
            "OFFSET", "WHERE", "HAVING", "WITH", "';'", 'r"\\)"', context=_context
        )
        if _token == "OFFSET":
            OFFSET = self._scan("OFFSET", context=_context)
            INT = self._scan("INT", context=_context)
            R.set_offset(int(INT))
            return True
        else:  # in ['WHERE', 'HAVING', 'WITH', "';'", 'r"\\)"']
            pass

    def where(self, S, _parent=None):
        _context = self.Context(_parent, self._scanner, "where", [S])
        _token = self._peek(
            "WHERE", "HAVING", "WITH", "';'", 'r"\\)"', context=_context
        )
        if _token == "WHERE":
            WHERE = self._scan("WHERE", context=_context)
            restriction = self.restriction(S, _context)
            S.set_where(restriction)
        else:  # in ['HAVING', 'WITH', "';'", 'r"\\)"']
            pass

    def restriction(self, S, _parent=None):
        _context = self.Context(_parent, self._scanner, "restriction", [S])
        rels_or = self.rels_or(S, _context)
        node = rels_or
        while (
            self._peek("','", 'r"\\)"', "HAVING", "WITH", "';'", context=_context)
            == "','"
        ):
            self._scan("','", context=_context)
            rels_or = self.rels_or(S, _context)
            node = And(node, rels_or)
        return node

    def rels_or(self, S, _parent=None):
        _context = self.Context(_parent, self._scanner, "rels_or", [S])
        rels_and = self.rels_and(S, _context)
        node = rels_and
        while (
            self._peek("OR", "','", 'r"\\)"', "HAVING", "WITH", "';'", context=_context)
            == "OR"
        ):
            OR = self._scan("OR", context=_context)
            rels_and = self.rels_and(S, _context)
            node = Or(node, rels_and)
        return node

    def rels_and(self, S, _parent=None):
        _context = self.Context(_parent, self._scanner, "rels_and", [S])
        rels_not = self.rels_not(S, _context)
        node = rels_not
        while (
            self._peek(
                "AND", "OR", "','", 'r"\\)"', "HAVING", "WITH", "';'", context=_context
            )
            == "AND"
        ):
            AND = self._scan("AND", context=_context)
            rels_not = self.rels_not(S, _context)
            node = And(node, rels_not)
        return node

    def rels_not(self, S, _parent=None):
        _context = self.Context(_parent, self._scanner, "rels_not", [S])
        _token = self._peek("NOT", 'r"\\("', "EXISTS", "VARIABLE", context=_context)
        if _token == "NOT":
            NOT = self._scan("NOT", context=_context)
            rel = self.rel(S, _context)
            return Not(rel)
        else:  # in ['r"\\("', 'EXISTS', 'VARIABLE']
            rel = self.rel(S, _context)
            return rel

    def rel(self, S, _parent=None):
        _context = self.Context(_parent, self._scanner, "rel", [S])
        _token = self._peek('r"\\("', "EXISTS", "VARIABLE", context=_context)
        if _token != 'r"\\("':
            rel_base = self.rel_base(S, _context)
            return rel_base
        else:  # == 'r"\\("'
            self._scan('r"\\("', context=_context)
            restriction = self.restriction(S, _context)
            self._scan('r"\\)"', context=_context)
            return restriction

    def rel_base(self, S, _parent=None):
        _context = self.Context(_parent, self._scanner, "rel_base", [S])
        _token = self._peek("EXISTS", "VARIABLE", context=_context)
        if _token == "VARIABLE":
            var = self.var(S, _context)
            opt_left = self.opt_left(S, _context)
            rtype = self.rtype(_context)
            rtype.append(var)
            rtype.set_optional(opt_left)
            expr = self.expr(S, _context)
            opt_right = self.opt_right(S, _context)
            rtype.append(expr)
            rtype.set_optional(opt_right)
            return rtype
        else:  # == 'EXISTS'
            EXISTS = self._scan("EXISTS", context=_context)
            self._scan('r"\\("', context=_context)
            restriction = self.restriction(S, _context)
            self._scan('r"\\)"', context=_context)
            return Exists(restriction)

    def rtype(self, _parent=None):
        _context = self.Context(_parent, self._scanner, "rtype", [])
        R_TYPE = self._scan("R_TYPE", context=_context)
        return Relation(R_TYPE)

    def opt_left(self, S, _parent=None):
        _context = self.Context(_parent, self._scanner, "opt_left", [S])
        _token = self._peek("QMARK", "R_TYPE", "CMP_OP", "'IN'", context=_context)
        if _token == "QMARK":
            QMARK = self._scan("QMARK", context=_context)
            return "left"
        else:  # in ['R_TYPE', 'CMP_OP', "'IN'"]
            pass

    def opt_right(self, S, _parent=None):
        _context = self.Context(_parent, self._scanner, "opt_right", [S])
        _token = self._peek(
            "QMARK",
            "AND",
            "OR",
            "','",
            'r"\\)"',
            "WITH",
            "';'",
            "HAVING",
            context=_context,
        )
        if _token == "QMARK":
            QMARK = self._scan("QMARK", context=_context)
            return "right"
        else:  # in ['AND', 'OR', "','", 'r"\\)"', 'WITH', "';'", 'HAVING']
            pass

    def logical_expr(self, S, _parent=None):
        _context = self.Context(_parent, self._scanner, "logical_expr", [S])
        exprs_or = self.exprs_or(S, _context)
        node = exprs_or
        while self._peek("','", 'r"\\)"', "WITH", "';'", context=_context) == "','":
            self._scan("','", context=_context)
            exprs_or = self.exprs_or(S, _context)
            node = And(node, exprs_or)
        return node

    def exprs_or(self, S, _parent=None):
        _context = self.Context(_parent, self._scanner, "exprs_or", [S])
        exprs_and = self.exprs_and(S, _context)
        node = exprs_and
        while (
            self._peek("OR", "','", 'r"\\)"', "WITH", "';'", context=_context) == "OR"
        ):
            OR = self._scan("OR", context=_context)
            exprs_and = self.exprs_and(S, _context)
            node = Or(node, exprs_and)
        return node

    def exprs_and(self, S, _parent=None):
        _context = self.Context(_parent, self._scanner, "exprs_and", [S])
        exprs_not = self.exprs_not(S, _context)
        node = exprs_not
        while (
            self._peek("AND", "OR", "','", 'r"\\)"', "WITH", "';'", context=_context)
            == "AND"
        ):
            AND = self._scan("AND", context=_context)
            exprs_not = self.exprs_not(S, _context)
            node = And(node, exprs_not)
        return node

    def exprs_not(self, S, _parent=None):
        _context = self.Context(_parent, self._scanner, "exprs_not", [S])
        _token = self._peek(
            "NOT",
            'r"\\("',
            "UNARY_OP",
            "NULL",
            "DATE",
            "DATETIME",
            "TRUE",
            "FALSE",
            "FLOAT",
            "INT",
            "STRING",
            "SUBSTITUTE",
            "VARIABLE",
            "E_TYPE",
            "FUNCTION",
            context=_context,
        )
        if _token == "NOT":
            NOT = self._scan("NOT", context=_context)
            balanced_expr = self.balanced_expr(S, _context)
            return Not(balanced_expr)
        else:
            balanced_expr = self.balanced_expr(S, _context)
            return balanced_expr

    def balanced_expr(self, S, _parent=None):
        _context = self.Context(_parent, self._scanner, "balanced_expr", [S])
        _token = self._peek(
            'r"\\("',
            "UNARY_OP",
            "NULL",
            "DATE",
            "DATETIME",
            "TRUE",
            "FALSE",
            "FLOAT",
            "INT",
            "STRING",
            "SUBSTITUTE",
            "VARIABLE",
            "E_TYPE",
            "FUNCTION",
            context=_context,
        )
        if _token == 'r"\\("':
            self._scan('r"\\("', context=_context)
            logical_expr = self.logical_expr(S, _context)
            self._scan('r"\\)"', context=_context)
            return logical_expr
        elif 1:
            expr_add = self.expr_add(S, _context)
            opt_left = self.opt_left(S, _context)
            expr_op = self.expr_op(S, _context)
            opt_right = self.opt_right(S, _context)
            expr_op.insert(0, expr_add)
            expr_op.set_optional(opt_left, opt_right)
            return expr_op
        else:
            raise runtime.SyntaxError(_token[0], "Could not match balanced_expr")

    def expr_op(self, S, _parent=None):
        _context = self.Context(_parent, self._scanner, "expr_op", [S])
        _token = self._peek("CMP_OP", "'IN'", context=_context)
        if _token == "CMP_OP":
            CMP_OP = self._scan("CMP_OP", context=_context)
            expr_add = self.expr_add(S, _context)
            return Comparison(CMP_OP.upper(), expr_add)
        else:  # == "'IN'"
            in_expr = self.in_expr(S, _context)
            return Comparison("=", in_expr)

    def variables(self, S, _parent=None):
        _context = self.Context(_parent, self._scanner, "variables", [S])
        vars = []
        var = self.var(S, _context)
        vars.append(var)
        while self._peek("','", "BEING", context=_context) == "','":
            self._scan("','", context=_context)
            var = self.var(S, _context)
            vars.append(var)
        return vars

    def decl_vars(self, R, _parent=None):
        _context = self.Context(_parent, self._scanner, "decl_vars", [R])
        E_TYPE = self._scan("E_TYPE", context=_context)
        var = self.var(R, _context)
        while (
            self._peek(
                "','",
                "R_TYPE",
                "QMARK",
                "WHERE",
                '":"',
                "CMP_OP",
                "HAVING",
                "'IN'",
                "';'",
                "POW_OP",
                "BEING",
                "WITH",
                "MUL_OP",
                'r"\\)"',
                "ADD_OP",
                "SORT_DESC",
                "SORT_ASC",
                "SORT_NULLSFIRST",
                "SORT_NULLSLAST",
                "GROUPBY",
                "ORDERBY",
                "LIMIT",
                "OFFSET",
                "AND",
                "OR",
                context=_context,
            )
            == "','"
        ):
            R.add_main_variable(E_TYPE, var)
            self._scan("','", context=_context)
            E_TYPE = self._scan("E_TYPE", context=_context)
            var = self.var(R, _context)
        R.add_main_variable(E_TYPE, var)

    def decl_rels(self, R, _parent=None):
        _context = self.Context(_parent, self._scanner, "decl_rels", [R])
        simple_rel = self.simple_rel(R, _context)
        while (
            self._peek(
                "','", "WHERE", "HAVING", "WITH", "';'", 'r"\\)"', context=_context
            )
            == "','"
        ):
            R.add_main_relation(simple_rel)
            self._scan("','", context=_context)
            simple_rel = self.simple_rel(R, _context)
        R.add_main_relation(simple_rel)

    def simple_rel(self, R, _parent=None):
        _context = self.Context(_parent, self._scanner, "simple_rel", [R])
        var = self.var(R, _context)
        R_TYPE = self._scan("R_TYPE", context=_context)
        e = Relation(R_TYPE)
        e.append(var)
        expr_add = self.expr_add(R, _context)
        e.append(Comparison("=", expr_add))
        return e

    def expr(self, S, _parent=None):
        _context = self.Context(_parent, self._scanner, "expr", [S])
        _token = self._peek(
            "CMP_OP",
            "UNARY_OP",
            'r"\\("',
            "NULL",
            "DATE",
            "DATETIME",
            "TRUE",
            "FALSE",
            "FLOAT",
            "INT",
            "STRING",
            "SUBSTITUTE",
            "VARIABLE",
            "E_TYPE",
            "FUNCTION",
            context=_context,
        )
        if _token == "CMP_OP":
            CMP_OP = self._scan("CMP_OP", context=_context)
            expr_add = self.expr_add(S, _context)
            return Comparison(CMP_OP.upper(), expr_add)
        else:
            expr_add = self.expr_add(S, _context)
            return Comparison("=", expr_add)

    def expr_add(self, S, _parent=None):
        _context = self.Context(_parent, self._scanner, "expr_add", [S])
        _token = self._peek(
            "UNARY_OP",
            'r"\\("',
            "NULL",
            "DATE",
            "DATETIME",
            "TRUE",
            "FALSE",
            "FLOAT",
            "INT",
            "STRING",
            "SUBSTITUTE",
            "VARIABLE",
            "E_TYPE",
            "FUNCTION",
            context=_context,
        )
        if _token != "UNARY_OP":
            expr_mul = self.expr_mul(S, _context)
            node = expr_mul
            while (
                self._peek(
                    "ADD_OP",
                    "QMARK",
                    'r"\\)"',
                    "','",
                    "SORT_DESC",
                    "SORT_ASC",
                    "SORT_NULLSFIRST",
                    "SORT_NULLSLAST",
                    "CMP_OP",
                    "R_TYPE",
                    "'IN'",
                    "GROUPBY",
                    "ORDERBY",
                    "WHERE",
                    "HAVING",
                    "LIMIT",
                    "OFFSET",
                    "WITH",
                    "';'",
                    "AND",
                    "OR",
                    context=_context,
                )
                == "ADD_OP"
            ):
                ADD_OP = self._scan("ADD_OP", context=_context)
                expr_mul = self.expr_mul(S, _context)
                node = MathExpression(ADD_OP, node, expr_mul)
            return node
        else:  # == 'UNARY_OP'
            UNARY_OP = self._scan("UNARY_OP", context=_context)
            expr_mul = self.expr_mul(S, _context)
            node = UnaryExpression(UNARY_OP, expr_mul)
            while (
                self._peek(
                    "ADD_OP",
                    "QMARK",
                    'r"\\)"',
                    "','",
                    "SORT_DESC",
                    "SORT_ASC",
                    "SORT_NULLSFIRST",
                    "SORT_NULLSLAST",
                    "CMP_OP",
                    "R_TYPE",
                    "'IN'",
                    "GROUPBY",
                    "ORDERBY",
                    "WHERE",
                    "HAVING",
                    "LIMIT",
                    "OFFSET",
                    "WITH",
                    "';'",
                    "AND",
                    "OR",
                    context=_context,
                )
                == "ADD_OP"
            ):
                ADD_OP = self._scan("ADD_OP", context=_context)
                expr_mul = self.expr_mul(S, _context)
                node = MathExpression(ADD_OP, node, expr_mul)
            return node

    def expr_mul(self, S, _parent=None):
        _context = self.Context(_parent, self._scanner, "expr_mul", [S])
        expr_pow = self.expr_pow(S, _context)
        node = expr_pow
        while (
            self._peek(
                "MUL_OP",
                "ADD_OP",
                "QMARK",
                'r"\\)"',
                "','",
                "SORT_DESC",
                "SORT_ASC",
                "SORT_NULLSFIRST",
                "SORT_NULLSLAST",
                "CMP_OP",
                "R_TYPE",
                "'IN'",
                "GROUPBY",
                "ORDERBY",
                "WHERE",
                "HAVING",
                "LIMIT",
                "OFFSET",
                "WITH",
                "';'",
                "AND",
                "OR",
                context=_context,
            )
            == "MUL_OP"
        ):
            MUL_OP = self._scan("MUL_OP", context=_context)
            expr_pow = self.expr_pow(S, _context)
            node = MathExpression(MUL_OP, node, expr_pow)
        return node

    def expr_pow(self, S, _parent=None):
        _context = self.Context(_parent, self._scanner, "expr_pow", [S])
        expr_base = self.expr_base(S, _context)
        node = expr_base
        while (
            self._peek(
                "POW_OP",
                "MUL_OP",
                "ADD_OP",
                "QMARK",
                'r"\\)"',
                "','",
                "SORT_DESC",
                "SORT_ASC",
                "SORT_NULLSFIRST",
                "SORT_NULLSLAST",
                "CMP_OP",
                "R_TYPE",
                "'IN'",
                "GROUPBY",
                "ORDERBY",
                "WHERE",
                "HAVING",
                "LIMIT",
                "OFFSET",
                "WITH",
                "';'",
                "AND",
                "OR",
                context=_context,
            )
            == "POW_OP"
        ):
            POW_OP = self._scan("POW_OP", context=_context)
            expr_base = self.expr_base(S, _context)
            node = MathExpression(MUL_OP, node, expr_base)
        return node

    def expr_base(self, S, _parent=None):
        _context = self.Context(_parent, self._scanner, "expr_base", [S])
        _token = self._peek(
            'r"\\("',
            "NULL",
            "DATE",
            "DATETIME",
            "TRUE",
            "FALSE",
            "FLOAT",
            "INT",
            "STRING",
            "SUBSTITUTE",
            "VARIABLE",
            "E_TYPE",
            "FUNCTION",
            context=_context,
        )
        if _token not in ['r"\\("', "VARIABLE", "E_TYPE", "FUNCTION"]:
            const = self.const(_context)
            return const
        elif _token == "VARIABLE":
            var = self.var(S, _context)
            return var
        elif _token == "E_TYPE":
            etype = self.etype(S, _context)
            return etype
        elif _token == "FUNCTION":
            func = self.func(S, _context)
            return func
        else:  # == 'r"\\("'
            self._scan('r"\\("', context=_context)
            expr_add = self.expr_add(S, _context)
            self._scan('r"\\)"', context=_context)
            return expr_add

    def func(self, S, _parent=None):
        _context = self.Context(_parent, self._scanner, "func", [S])
        FUNCTION = self._scan("FUNCTION", context=_context)
        self._scan('r"\\("', context=_context)
        F = Function(FUNCTION)
        if (
            self._peek(
                "UNARY_OP",
                'r"\\)"',
                'r"\\("',
                "NULL",
                "DATE",
                "DATETIME",
                "TRUE",
                "FALSE",
                "FLOAT",
                "INT",
                "STRING",
                "SUBSTITUTE",
                "VARIABLE",
                "E_TYPE",
                "FUNCTION",
                context=_context,
            )
            != 'r"\\)"'
        ):
            expr_add = self.expr_add(S, _context)
            while (
                self._peek(
                    "','",
                    "QMARK",
                    'r"\\)"',
                    "SORT_DESC",
                    "SORT_ASC",
                    "SORT_NULLSFIRST",
                    "SORT_NULLSLAST",
                    "CMP_OP",
                    "R_TYPE",
                    "'IN'",
                    "GROUPBY",
                    "ORDERBY",
                    "WHERE",
                    "HAVING",
                    "LIMIT",
                    "OFFSET",
                    "WITH",
                    "';'",
                    "AND",
                    "OR",
                    context=_context,
                )
                == "','"
            ):
                F.append(expr_add)
                self._scan("','", context=_context)
                expr_add = self.expr_add(S, _context)
            F.append(expr_add)
        self._scan('r"\\)"', context=_context)
        return F

    def in_expr(self, S, _parent=None):
        _context = self.Context(_parent, self._scanner, "in_expr", [S])
        self._scan("'IN'", context=_context)
        self._scan('r"\\("', context=_context)
        F = Function("IN")
        if (
            self._peek(
                "UNARY_OP",
                'r"\\)"',
                'r"\\("',
                "NULL",
                "DATE",
                "DATETIME",
                "TRUE",
                "FALSE",
                "FLOAT",
                "INT",
                "STRING",
                "SUBSTITUTE",
                "VARIABLE",
                "E_TYPE",
                "FUNCTION",
                context=_context,
            )
            != 'r"\\)"'
        ):
            expr_add = self.expr_add(S, _context)
            while (
                self._peek(
                    "','",
                    "QMARK",
                    'r"\\)"',
                    "SORT_DESC",
                    "SORT_ASC",
                    "SORT_NULLSFIRST",
                    "SORT_NULLSLAST",
                    "CMP_OP",
                    "R_TYPE",
                    "'IN'",
                    "GROUPBY",
                    "ORDERBY",
                    "WHERE",
                    "HAVING",
                    "LIMIT",
                    "OFFSET",
                    "WITH",
                    "';'",
                    "AND",
                    "OR",
                    context=_context,
                )
                == "','"
            ):
                F.append(expr_add)
                self._scan("','", context=_context)
                expr_add = self.expr_add(S, _context)
            F.append(expr_add)
        self._scan('r"\\)"', context=_context)
        return F

    def var(self, S, _parent=None):
        _context = self.Context(_parent, self._scanner, "var", [S])
        VARIABLE = self._scan("VARIABLE", context=_context)
        return VariableRef(S.get_variable(VARIABLE))

    def etype(self, S, _parent=None):
        _context = self.Context(_parent, self._scanner, "etype", [S])
        E_TYPE = self._scan("E_TYPE", context=_context)
        return S.get_etype(E_TYPE)

    def const(self, _parent=None):
        _context = self.Context(_parent, self._scanner, "const", [])
        _token = self._peek(
            "NULL",
            "DATE",
            "DATETIME",
            "TRUE",
            "FALSE",
            "FLOAT",
            "INT",
            "STRING",
            "SUBSTITUTE",
            context=_context,
        )
        if _token == "NULL":
            NULL = self._scan("NULL", context=_context)
            return Constant(None, None)
        elif _token == "DATE":
            DATE = self._scan("DATE", context=_context)
            return Constant(DATE.upper(), "Date")
        elif _token == "DATETIME":
            DATETIME = self._scan("DATETIME", context=_context)
            return Constant(DATETIME.upper(), "Datetime")
        elif _token == "TRUE":
            TRUE = self._scan("TRUE", context=_context)
            return Constant(True, "Boolean")
        elif _token == "FALSE":
            FALSE = self._scan("FALSE", context=_context)
            return Constant(False, "Boolean")
        elif _token == "FLOAT":
            FLOAT = self._scan("FLOAT", context=_context)
            return Constant(float(FLOAT), "Float")
        elif _token == "INT":
            INT = self._scan("INT", context=_context)
            return Constant(int(INT), "Int")
        elif _token == "STRING":
            STRING = self._scan("STRING", context=_context)
            return Constant(unquote(STRING), "String")
        else:  # == 'SUBSTITUTE'
            SUBSTITUTE = self._scan("SUBSTITUTE", context=_context)
            return Constant(SUBSTITUTE[2:-2], "Substitute")


def parse(rule, text):
    P = Hercule(HerculeScanner(text))
    return runtime.wrap_error_reporter(P, rule)


# End -- grammar generated by Yapps


from warnings import warn
from rql.stmts import Union, Select, Delete, Insert, Set
from rql.nodes import (
    SubQuery,
    And,
    Or,
    Not,
    Exists,
    Relation,
    Comparison,
    MathExpression,
    UnaryExpression,
    Function,
    Constant,
    VariableRef,
    SortTerm,
)


def unquote(string):
    """Remove quotes from a string."""
    if string.startswith('"'):
        return string[1:-1].replace("\\\\", "\\").replace('\\"', '"')
    elif string.startswith("'"):
        return string[1:-1].replace("\\\\", "\\").replace("\\'", "'")



# copyright 2004-2021 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of rql.
#
# rql is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# rql is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with rql. If not, see <http://www.gnu.org/licenses/>.
"""Base classes for RQL syntax tree nodes.

Note: this module uses __slots__ to limit memory usage.
"""


from rql.utils import VisitableMixIn
from typing import (
    Tuple,
    List,
    Any,
    Optional,
    TYPE_CHECKING,
    Iterator,
    Union as Union_,
    Dict,
    Type,
    TypeVar,
    Iterable,
    Sequence,
)

__docformat__: str = "restructuredtext en"


if TYPE_CHECKING:
    import rql


_Y = TypeVar("_Y", bound="BaseNode")


class BaseNode(VisitableMixIn):
    __slots__: Iterable[str] = ("parent",)

    if TYPE_CHECKING:
        parent: Optional["BaseNode"]

        # the only class that doesn't have children is ScopeNode but it
        # doesn't seems to be used explicitely anywhere and ALL its children
        # have a "children" attribute, so we take the decision to make
        # everything simplier and, for typing, pretend that ALL nodes have
        # children
        # also ScopeNode really seems to be an abstract base class
        # note: ALL subclass of BaseNode (except ScopeNode) also have children
        children: Sequence["BaseNode"]

        # another option would be to define a new type like this:
        # NodeWithChildren = Union_[Node, LeafNode, SubQuery, Set, Insert, Delete,
        #                           Exists]
        # or create a protocol in the same fashion

    def __str__(self) -> str:
        s: str = self.as_string()
        return s

    def as_string(self, kwargs: Dict = None) -> str:
        """Return the tree as an encoded rql string."""
        raise NotImplementedError()

    def initargs(self, stmt: Optional["rql.stmts.Statement"]) -> Tuple[Any, ...]:
        """Return list of arguments to give to __init__ to clone this node.

        I don't use __getinitargs__ because I'm not sure it should interfer with
        copy/pickle
        """
        return ()

    # pyannotate suggested rql.stmts.Union as returned type, samples=2
    # but it sound like we should return a Node here
    @property
    def root(self) -> Optional["BaseNode"]:
        """Return the root node of the tree"""
        if self.parent is not None:
            return self.parent.root
        return None  # should we return None or something else?

    @property
    def stmt(self) -> Optional["rql.stmts.Statement"]:
        """Return the Select node to which this node belong"""
        if self.parent is not None:
            return self.parent.stmt
        return None  # should we return None or something else?

    @property
    def scope(self) -> Optional[Union_["rql.stmts.Statement", "BaseNode"]]:
        """Return the scope node to which this node belong (eg Select or Exists
        node)
        """
        if self.parent is not None:
            return self.parent.scope
        return None  # should we return None or something else?

    def get_nodes(self, klass: Type[_Y]) -> List[_Y]:
        """Return the list of nodes of a given class in the subtree.

        :type klass: a node class (Relation, Constant, etc.)
        :param klass: the class of nodes to return

        :rtype: list
        """
        stack: List[Any] = [self]
        result = []
        while stack:
            node = stack.pop()
            if isinstance(node, klass):
                result.append(node)
            else:
                stack += node.children
        return result

    def iget_nodes(self, klass: Type[_Y]) -> Iterator[_Y]:
        """Return an iterator over nodes of a given class in the subtree.

        :type klass: a node class (Relation, Constant, etc.)
        :param klass: the class of nodes to return

        :rtype: iterator
        """
        stack: List[Any] = [self]
        while stack:
            node = stack.pop()
            if isinstance(node, klass):
                yield node
            else:
                stack += node.children

    # Argument 1 to "is_equivalent" of "BaseNode" has incompatible type "IsOperator";
    # expected "BaseNode"  [arg-type]
    # Argument 2 to "is_equivalent" of "BaseNode" has incompatible type "IsOperator";
    # expected "BaseNode"  [arg-type]
    # def is_equivalent(self, other: "BaseNode") -> bool:
    # Any here really means: any children of BaseNode
    def is_equivalent(self: Any, other: Any) -> bool:
        if other.__class__ is not self.__class__:
            return False
        for i, child in enumerate(self.children):
            try:
                if not child.is_equivalent(other.children[i]):
                    return False
            except IndexError:
                return False
        return True

    def copy(self, stmt: Optional["rql.stmts.Statement"] = None) -> "BaseNode":
        raise NotImplementedError()

    def replace(
        self, old_child: "BaseNode", new_child: "BaseNode"
    ) -> Tuple["BaseNode", "BaseNode", Optional[int]]:
        raise NotImplementedError()


class Node(BaseNode):
    """Class for nodes of the tree which may have children (almost all...)"""

    __slots__: Iterable[str] = ("children",)

    def __init__(self):
        self.parent: Optional[BaseNode] = None
        self.children: List[BaseNode] = []

    def append(self, child: BaseNode):
        """add a node to children"""
        self.children.append(child)
        child.parent = self

    def remove(
        self, child: BaseNode
    ) -> Tuple[BaseNode, Optional[BaseNode], Optional[int]]:
        """Remove a child node. Return the removed node, its old parent and
        index in the children list.
        """
        index: int = self.children.index(child)
        del self.children[index]
        parent = child.parent
        child.parent = None
        return child, parent, index

    def insert(self, index: int, child: BaseNode):
        """insert a child node"""
        self.children.insert(index, child)
        child.parent = self

    def replace(
        self, old_child: BaseNode, new_child: BaseNode
    ) -> Tuple[BaseNode, BaseNode, Optional[int]]:
        """replace a child node with another"""
        i = self.children.index(old_child)
        self.children.pop(i)
        self.children.insert(i, new_child)
        new_child.parent = self
        return old_child, self, i

    def copy(self, stmt: Optional["rql.stmts.Statement"] = None) -> "Node":
        """Create and return a copy of this node and its descendant.

        stmt is the root node, which should be use to get new variables
        """
        new: Node = self.__class__(*self.initargs(stmt))
        for child in self.children:
            new.append(child.copy(stmt))
        return new


class BinaryNode(Node):
    __slots__: Iterable[str] = ()

    def __init__(
        self,
        lhs: Optional["BaseNode"] = None,
        rhs: Optional["BaseNode"] = None,
    ):
        Node.__init__(self)
        if lhs is not None:
            self.append(lhs)
        if rhs is not None:
            self.append(rhs)

    def remove(
        self, child: "BaseNode"
    ) -> Tuple["BaseNode", Optional["BaseNode"], Optional[int]]:
        """Remove the child and replace this node with the other child."""
        index: int = self.children.index(child)
        if self.parent is not None:
            return self.parent.replace(self, self.children[not index])
        return child, None, index

    def get_parts(self) -> Tuple["BaseNode", "BaseNode"]:
        """Return the left hand side and the right hand side of this node."""
        return self.children[0], self.children[1]


class LeafNode(BaseNode):
    """Class optimized for leaf nodes."""

    __slots__: Iterable[str] = ()

    # We can't do that because of a bug in mypy see
    # https://github.com/python/mypy/issues/4125#issuecomment-337187251
    # @property
    # def children(self) -> Sequence["BaseNode"]:
    @property
    def children(self):
        return ()

    def copy(
        self: "BaseNode", stmt: Optional["rql.stmts.Statement"] = None
    ) -> "BaseNode":
        """Create and return a copy of this node and its descendant.

        stmt is the root node, which should be use to get new variables.
        """
        return self.__class__(*self.initargs(stmt))



# copyright 2004-2021 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of rql.
#
# rql is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# rql is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with rql. If not, see <http://www.gnu.org/licenses/>.
"""RQL library (implementation independant)."""

from math import log

import sys
import threading

if sys.version_info < (3, 10):
    from importlib_metadata import version
else:
    from importlib.metadata import version

from io import StringIO

from rql._exceptions import (  # noqa
    UsesReservedWord,
    RQLSyntaxError,
    RQLException,
    MissingType,
    UsesReservedWord,
    RQLSyntaxError,
    TypeResolverException,
    BadRQLQuery,
    CoercionError,
)
from rql.analyze import ETypeResolver
from typing import TYPE_CHECKING, Optional, Dict, List, Any, Set, cast

__docformat__: str = "restructuredtext en"
__version__: str = version("rql")

if TYPE_CHECKING:
    import rql

# REQUIRED_TYPES = ['String', 'Float', 'Int', 'Boolean', 'Date']


class RQLHelper:
    """Helper class for RQL handling

    give access to methods for :
      - parsing RQL strings
      - variables type resolving
      - comparison of two queries
    """

    def __init__(
        self,
        schema: "rql.interfaces.ISchema",
        uid_func_mapping: "rql.analyze.UIDFuncMapping" = None,
        special_relations: Dict[str, str] = {},
        backend: Optional[Any] = None,
    ) -> None:
        # chech schema
        # for e_type in REQUIRED_TYPES:
        #    if not schema.has_entity(e_type):
        #        raise MissingType(e_type)
        # create helpers
        from rql.stcheck import RQLSTChecker, RQLSTAnnotator

        if uid_func_mapping:
            for key in uid_func_mapping:
                special_relations[key] = "uid"
        self._checker = RQLSTChecker(schema, special_relations, backend)
        self._annotator = RQLSTAnnotator(schema, special_relations)
        self._analyser_lock = threading.Lock()
        self._analyser = ETypeResolver(schema, uid_func_mapping)
        # IgnoreTypeRestriction analyser
        from rql.analyze import ETypeResolverIgnoreTypeRestriction

        self._itr_analyser_lock = threading.Lock()
        self._itr_analyser = ETypeResolverIgnoreTypeRestriction(
            schema, uid_func_mapping
        )
        self.set_schema(schema)

    def set_schema(self, schema: "rql.interfaces.ISchema") -> None:
        from rql.utils import is_keyword

        for etype in schema.entities():
            etype = str(etype)
            if is_keyword(etype) or etype.capitalize() == "Any":
                raise UsesReservedWord(etype)
        for rtype in schema.relations():
            rtype = str(rtype)
            if is_keyword(rtype):
                raise UsesReservedWord(rtype)
        self._checker.schema = schema
        self._annotator.schema = schema
        self._analyser.set_schema(schema)

    def get_backend(self) -> Any:
        return self._checker.backend

    def set_backend(self, backend: Any) -> None:
        self._checker.backend = backend

    backend = property(get_backend, set_backend)

    def parse(self, rqlstring: str, annotate: bool = True) -> "rql.stmts.AnyStatement":
        """Return a syntax tree created from a RQL string."""
        rqlst = parse(rqlstring, False)
        self._checker.check(rqlst)
        if annotate:
            self.annotate(rqlst)
        rqlst.schema = self._annotator.schema
        return rqlst

    def annotate(self, rqlst: "rql.stmts.Statement") -> None:
        self._annotator.annotate(rqlst)

    def compute_solutions(
        self,
        rqlst: "rql.stmts.AnyStatement",
        uid_func_mapping: "rql.analyze.UIDFuncMapping" = None,
        kwargs: Optional[Dict] = None,
        debug: bool = False,
    ) -> Set:
        """Set solutions for variables of the syntax tree.

        Each solution is a dictionary with variable's name as key and
        variable's type as value.
        """
        with self._analyser_lock:
            return self._analyser.visit(rqlst, uid_func_mapping, kwargs, debug)

    def compute_all_solutions(
        self, rqlst, uid_func_mapping=None, kwargs=None, debug=False
    ):
        """compute syntax tree solutions with all types restriction (eg
        is/instance_of relations) ignored
        """
        with self._itr_analyser_lock:
            self._itr_analyser.visit(rqlst, uid_func_mapping, kwargs, debug)

    def simplify(self, rqlst: "rql.stmts.Union") -> None:
        """Simplify `rqlst` by rewriting non-final variables associated to a const
        node (if annotator say we can...)

        The tree is modified in-place.
        """
        # print('simplify', rqlst.as_string())
        # seems like only Union as TYPE attribute, I should double check
        if rqlst.TYPE == "select":
            for select in rqlst.children:
                self._simplify(cast("rql.stmts.Select", select))

    def _simplify(self, select: "rql.stmts.Select") -> None:
        # recurse on subqueries first
        for subquery in select.with_:
            for subselect in subquery.query.children:
                self._simplify(cast("rql.stmts.Select", subselect))
        rewritten = False
        for var in list(select.defined_vars.values()):
            stinfo = var.stinfo
            if (
                stinfo["constnode"]
                and not stinfo.get("blocsimplification")
                and not var.has_attribute_or_function_var_references()
            ):
                uidrel = stinfo["uidrel"]
                var = uidrel.children[0].variable
                vconsts: List["rql.nodes.Constant"] = []
                rhs = uidrel.children[1].children[0]
                for vref in var.references():
                    rel = vref.relation()
                    if rel is None:
                        term = vref
                        while term.parent is not select:
                            term = term.parent
                        if any(term.is_equivalent(t) for t in select.selection):
                            rhs = copy_uid_node(select, rhs, vconsts)
                            if vref is term:
                                index = next(
                                    i
                                    for i, var in enumerate(select.selection)
                                    if vref.is_equivalent(var)
                                )
                                select.selection[index] = rhs
                                rhs.parent = select
                            else:
                                vref.parent.replace(vref, rhs)
                        elif any(term.is_equivalent(o) for o in select.orderby):
                            # remove from orderby
                            select.remove(term)
                        elif not select.having:
                            # remove from groupby if no HAVING clause
                            select.remove(term)
                        else:
                            rhs = copy_uid_node(select, rhs, vconsts)
                            # "Iterable[Any]" has no attribute "index"  [attr-defined]
                            select.groupby = cast(List, select.groupby)
                            select.groupby[select.groupby.index(vref)] = rhs
                            rhs.parent = select
                    elif rel is uidrel:
                        uidrel.parent.remove(uidrel)
                    elif rel.is_types_restriction():
                        stinfo["typerel"] = None
                        rel.parent.remove(rel)
                    else:
                        rhs = copy_uid_node(select, rhs, vconsts)
                        vref.parent.replace(vref, rhs)
                del select.defined_vars[var.name]
                stinfo["uidrel"] = None
                rewritten = True
                if vconsts:
                    select.stinfo["rewritten"][var.name] = vconsts
        if rewritten and select.solutions:
            select.clean_solutions()


def copy_uid_node(
    select: "rql.stmts.Select",
    node: "rql.nodes.Constant",
    vconsts: List["rql.nodes.Constant"],
) -> "rql.nodes.Constant":
    # Incompatible types in assignment
    #  (expression has type "BaseNode", variable has type "Constant")
    node = node.copy(select)  # type:ignore[assignment]
    node.uid = True
    vconsts.append(node)
    return node


def parse(rqlstring: str, print_errors: bool = True) -> "rql.stmts.AnyStatement":
    """Return a syntax tree created from a RQL string."""
    from yapps.runtime import print_error, SyntaxError, NoMoreTokens
    from rql.parser import Hercule, HerculeScanner

    # make sure rql string ends with a semi-colon
    rqlstring = rqlstring.strip()
    if rqlstring and not rqlstring.endswith(";"):
        rqlstring += ";"
    # parse the RQL string
    parser = Hercule(HerculeScanner(rqlstring))
    try:
        return parser.goal()
    except SyntaxError as ex:
        if not print_errors:
            if ex.pos is not None:
                multi_lines_rql = rqlstring.splitlines()
                nb_lines = len(multi_lines_rql)
                if nb_lines > 5:
                    width = log(nb_lines, 10) + 1
                    template = " %%%ii: %%s" % width
                    rqlstring = "\n".join(
                        template % (idx + 1, line)
                        for idx, line in enumerate(multi_lines_rql)
                    )

                padding = 30
                _, line, column = ex.pos

                rql_line = multi_lines_rql[line - 1]

                if len(rql_line) <= (padding * 2):
                    around = rql_line
                    here_left_padding = len("around: ") + column
                elif column < padding:
                    around = rql_line[: padding * 2] + "..."
                    here_left_padding = len("around: ") + column
                elif (len(rql_line) - padding) < column:
                    around = "..." + rql_line[-(padding * 2) :]
                    here_left_padding = len("around: ...") + (
                        column - (len(rql_line) - (padding * 2))
                    )
                else:
                    around = (
                        "..." + rql_line[column - padding : column + padding] + "..."
                    )
                    here_left_padding = len("around: ...") + padding

                here = " " * (here_left_padding - 1) + "^ here"
                msg = "%s in:\n%s\n\nat line %s and column %s\naround: %s\n%s" % (
                    ex.msg,
                    rqlstring,
                    line,
                    column,
                    around,
                    here,
                )
            else:
                msg = f"{rqlstring}\n{ex.msg}"
            exc = RQLSyntaxError(msg)
            exc.__traceback__ = sys.exc_info()[-1]
            raise exc
        # try to get error message from yapps
        try:
            out = sys.stderr
            sys.stderr = stream = StringIO()
            try:
                print_error(ex, parser._scanner)
            finally:
                sys.stderr = out
            exc = RQLSyntaxError(stream.getvalue())
            exc.__traceback__ = sys.exc_info()[-1]
            raise exc
        except ImportError:  # duh?
            sys.stderr = out
            exc = RQLSyntaxError(
                "Syntax Error",
                ex.msg,
                "on line",
                1 + out.count("\n", 0, ex.pos),  # type:ignore
            )
            exc.__traceback__ = sys.exc_info()[-1]
            raise exc
    except NoMoreTokens:
        msg = "Could not complete parsing; stopped around here: \n%s"
        exc = RQLSyntaxError(msg % parser._scanner)
        exc.__traceback__ = sys.exc_info()[-1]
        raise exc



# copyright 2004-2021 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of rql.
#
# rql is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# rql is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with rql. If not, see <http://www.gnu.org/licenses/>.

import unittest
from logilab.common.testlib import TestCase

from rql import parse
from rql.nodes import Exists


class RQLUndoTestCase(TestCase):
    def test_selected(self):
        rqlst = parse("Person X")
        orig = rqlst.as_string()
        rqlst.save_state()
        select = rqlst.children[0]
        var = select.make_variable()
        select.remove_selected(select.selection[0])
        select.add_selected(var)
        # check operations
        self.assertEqual(rqlst.as_string(), f"Any {var.name} WHERE X is Person")
        # check references before recovering
        rqlst.check_references()
        rqlst.recover()
        # check equivalence after recovering
        self.assertEqual(rqlst.as_string(), orig)
        # check references after recovering
        rqlst.check_references()

    def test_selected3(self):
        rqlst = parse("Any lower(N) WHERE X is Person, X name N")
        orig = rqlst.as_string()
        rqlst.save_state()
        select = rqlst.children[0]
        var = select.make_variable()
        select.remove_selected(select.selection[0])
        select.add_selected(var)
        # check operations
        self.assertEqual(
            rqlst.as_string(), f"Any {var.name} WHERE X is Person, X name N"
        )
        # check references before recovering
        rqlst.check_references()
        rqlst.recover()
        # check equivalence after recovering
        self.assertEqual(rqlst.as_string(), orig)
        # check references after recovering
        rqlst.check_references()

    def test_undefine_1(self):
        rqlst = parse("Person X, Y WHERE X travaille_pour Y")
        orig = rqlst.as_string()
        rqlst.save_state()
        rqlst.children[0].undefine_variable(rqlst.children[0].defined_vars["Y"])
        # check operations
        self.assertEqual(rqlst.as_string(), "Any X WHERE X is Person")
        # check references before recovering
        rqlst.check_references()
        rqlst.recover()
        # check equivalence
        self.assertEqual(rqlst.as_string(), orig)
        # check references after recovering
        rqlst.check_references()

    def test_undefine_2(self):
        rqlst = parse("Person X")
        orig = rqlst.as_string()
        rqlst.save_state()
        rqlst.children[0].undefine_variable(rqlst.children[0].defined_vars["X"])
        var = rqlst.children[0].make_variable()
        rqlst.children[0].add_selected(var)
        # check operations
        self.assertEqual(rqlst.as_string(), "Any A")
        # check references before recovering
        rqlst.check_references()
        rqlst.recover()
        # check equivalence
        self.assertEqual(rqlst.as_string(), orig)
        # check references after recovering
        rqlst.check_references()

    def test_remove_exists(self):
        rql_request = (
            "Any U,COUNT(P) GROUPBY U WHERE U is CWUser, "
            "P? patch_reviewer U, EXISTS(P in_state S AND "
            'S name "pouet")'
        )
        rqlst = parse(rql_request).children[0]
        orig = rqlst.as_string()
        rqlst.save_state()
        n = [r for r in rqlst.get_nodes(Exists)][0].query
        rqlst.remove_node(n)
        # check operations
        self.assertEqual(
            rqlst.as_string(),
            "Any U,COUNT(P) GROUPBY U WHERE U is CWUser, P? patch_reviewer U",
        )
        # check references before recovering
        rqlst.check_references()
        rqlst.recover()
        # check equivalence
        self.assertEqual(rqlst.as_string(), orig)
        # check references after recovering
        rqlst.check_references()


if __name__ == "__main__":
    unittest.main()



# copyright 2004-2021 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of rql.
#
# rql is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# rql is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with rql. If not, see <http://www.gnu.org/licenses/>.

import unittest

from rql import RQLHelper, BadRQLQuery, stmts, nodes
from unittest_analyze import DummySchema


BAD_QUERIES = (
    "Any X, Y GROUPBY X",
    "Any X WHERE X name Person",
    "Any X WHERE X name nofunction(Y)",
    "Any X WHERE X name nofunction(Y)",
    'Any Y WHERE X name "toto"',
    'Any X WHERE X noattr "toto"',
    "Any X WHERE X is NonExistant",
    'Any UPPER(Y) WHERE X name "toto"',
    "Any C ORDERBY N where C located P, P eid %(x)s",
    "Any X, MAX(COUNT(B)) GROUPBY X WHERE B concerns X;",
    '(Any X WHERE X nom "toto") UNION (Any X,F WHERE X firstname F);',
    "Any X, X/Z WHERE X is Person; WITH Y BEING (Any SUM(X) WHERE X is Person)",
    'Any X WHERE X name "Toto", P is Person',
    "Any X WHERE X eid 0, X eid 1",
    # DISTINCT+ORDERBY tests ###################################################
    # cant sort on Y, B <- work_for X is multivalued
    "DISTINCT Any X ORDERBY Y WHERE B work_for X, B name Y",
    # cant sort on PN, there may be different PF values for the same PN value
    # XXX untrue if PF or PN is marked as unique
    "DISTINCT Any PF ORDERBY PN WHERE P firstname PF, P name PN",
    # cant sort on XN, there may be different PF values for the same PF value
    "DISTINCT Any PF ORDERBY X WHERE P work_for X, P firstname PF",
    "DISTINCT Any PF ORDERBY XN WHERE P work_for X, P firstname PF, X name XN",
)

OK_QUERIES = (
    "(Any N,COUNT(X) GROUPBY N WHERE X name N)"
    " UNION "
    "(Any N,COUNT(X) GROUPBY N WHERE X firstname N)",
    "DISTINCT Any X, MAX(Y) GROUPBY X WHERE X is Person, Y is Company",
    # DISTINCT+ORDERBY tests ###################################################
    # sorting allowed since order variable reachable from a selected
    # variable with only ?1 cardinality
    "DISTINCT Any P ORDERBY PN WHERE P work_for X, P name PN",
    "DISTINCT Any P ORDERBY XN WHERE P work_for X, X name XN",
    "Any X WHERE X eid > 0, X eid < 42",
    "Any X WHERE X eid 1, X eid < 42",
    "Any X WHERE X number CAST(Int, Y), X name Y",
    "SET X number CAST(Int, Y) WHERE X name Y",
)


class CheckClassTest(unittest.TestCase):
    """check wrong queries are correctly detected"""

    def setUp(self):
        helper = RQLHelper(DummySchema(), None, {"eid": "uid"})
        self.parse = helper.parse
        self.simplify = helper.simplify

    def _test(self, rql):
        try:
            self.assertRaises(BadRQLQuery, self.parse, rql)
        except Exception:
            print(rql)
            raise

    def test_raise(self):
        for rql in BAD_QUERIES:
            with self.subTest(rql=rql):
                self._test(rql)

    def test_ok(self):
        for rql in OK_QUERIES:
            with self.subTest(rql=rql):
                self.parse(rql)

    def _test_rewrite(self, rql, expected):
        rqlst = self.parse(rql)
        self.simplify(rqlst)
        self.assertEqual(rqlst.as_string(), expected)

    def test_rewrite(self):
        for rql, expected in (
            ("Person X", "Any X WHERE X is Person"),
            (
                "Any X WHERE X eid IN (12), X name 'toto'",
                'Any X WHERE X eid 12, X name "toto"',
            ),
            ("Any X WHERE X work_for Y, Y eid 12", "Any X WHERE X work_for 12"),
            ("Any X WHERE Y work_for X, Y eid 12", "Any X WHERE 12 work_for X"),
            ("Any X WHERE X work_for Y, Y eid IN (12)", "Any X WHERE X work_for 12"),
            (
                "Any X ORDERBY Y WHERE X work_for Y, Y eid IN (12)",
                "Any X WHERE X work_for 12",
            ),
            ("Any X WHERE X eid 12", "Any 12"),
            ("Any X WHERE X is Person, X eid 12", "Any 12"),
            (
                "Any X,Y WHERE X eid 0, Y eid 1, X work_for Y",
                "Any 0,1 WHERE 0 work_for 1",
            ),
            # test symmetric OR rewrite
            (
                "DISTINCT Any P WHERE P connait S OR S connait P, S name 'chouette'",
                'DISTINCT Any P WHERE P connait S, S name "chouette"',
            ),
            # queries that should not be rewritten
            ("DELETE Person X WHERE X eid 12", "DELETE Person X WHERE X eid 12"),
            (
                "Any X WHERE X work_for Y, Y eid IN (12, 13)",
                "Any X WHERE X work_for Y, Y eid IN(12, 13)",
            ),
            (
                "Any X WHERE X work_for Y, NOT Y eid 12",
                "Any X WHERE X work_for Y, NOT Y eid 12",
            ),
            ("Any X WHERE NOT X eid 12", "Any X WHERE NOT X eid 12"),
            ("Any N WHERE X eid 12, X name N", "Any N WHERE X eid 12, X name N"),
            ("Any X WHERE X eid > 12", "Any X WHERE X eid > 12"),
            (
                "Any X WHERE X eid 12, X connait P?, X work_for Y",
                "Any X WHERE X eid 12, X connait P?, X work_for Y",
            ),
            (
                "Any X WHERE X eid 12, P? connait X",
                "Any X WHERE X eid 12, P? connait X",
            ),
            (
                "Any X WHERE X firstname 'lulu',"
                "EXISTS (X owned_by U, U name 'lulufanclub' OR U name 'managers');",
                'Any X WHERE X firstname "lulu", '
                'EXISTS(X owned_by U, (U name "lulufanclub") OR (U name "managers"))',
            ),
            (
                'Any X WHERE X eid 12, EXISTS(X name "hop" OR X work_for Y?)',
                'Any 12 WHERE EXISTS((A name "hop") OR (A work_for Y?), 12 identity A)',
            ),
            (
                "(Any X WHERE X eid 12) UNION (Any X ORDERBY X WHERE X eid 13)",
                "(Any 12) UNION (Any 13)",
            ),
            (
                "Any X WITH X BEING (Any X WHERE X eid 12)",
                "Any X WITH X BEING (Any 12)",
            ),
            (
                "Any X GROUPBY X WHERE X eid 12, X connait Y HAVING COUNT(Y) > 10",
                "Any X GROUPBY X WHERE X eid 12, X connait Y HAVING COUNT(Y) > 10",
            ),
            # A eid 12 can be removed since the type analyzer checked its existence
            ("Any X WHERE A eid 12, X connait Y", "Any X WHERE X connait Y"),
            (
                "Any X WHERE EXISTS(X work_for Y, Y eid 12) OR X eid 12",
                "Any X WHERE (EXISTS(X work_for 12)) OR (X eid 12)",
            ),
            (
                "Any X WHERE EXISTS(X work_for Y, Y eid IN (12)) OR X eid IN (12)",
                "Any X WHERE (EXISTS(X work_for 12)) OR (X eid 12)",
            ),
            ("Any X WHERE NOT EXISTS(X eid 1)", "Any X WHERE NOT EXISTS(X eid 1)"),
        ):
            with self.subTest(rql=rql):
                self._test_rewrite(rql, expected)

    def test_subquery_graphdict(self):
        # test two things:
        # * we get graph information from subquery
        # * we see that we can sort on VCS (eg we have a unique value path from VF to VCD)
        rqlst = self.parse(
            (
                "DISTINCT Any VF ORDERBY VCD DESC WHERE "
                'VC work_for S, S name "draft" '
                "WITH VF, VC, VCD BEING (Any VF, MAX(VC), VCD GROUPBY VF, VCD "
                "                        WHERE VC connait VF, VC creation_date VCD)"
            )
        )
        self.assertEqual(
            rqlst.children[0].vargraph,
            {
                "VCD": ["VC"],
                "VF": ["VC"],
                "S": ["VC"],
                "VC": ["S", "VF", "VCD"],
                ("VC", "S"): "work_for",
                ("VC", "VF"): "connait",
                ("VC", "VCD"): "creation_date",
            },
        )
        self.assertEqual(rqlst.children[0].aggregated, set(("VC",)))


class CopyTest(unittest.TestCase):
    def setUp(self):
        helper = RQLHelper(DummySchema(), None, {"eid": "uid"})
        self.parse = helper.parse
        self.simplify = helper.simplify
        self.annotate = helper.annotate

    def test_copy_exists(self):
        tree = self.parse(
            "Any X WHERE X firstname 'lulu',"
            "EXISTS (X owned_by U, U work_for G, G name 'lulufanclub' OR G name 'managers');"
        )
        self.simplify(tree)
        copy = tree.copy()
        exists = copy.get_nodes(nodes.Exists)[0]
        self.assertIs(exists.children[0].parent, exists)
        self.assertTrue(exists.parent)

    def test_copy_internals(self):
        root = self.parse(
            "Any X,U WHERE C owned_by U, NOT X owned_by U, X eid 1, C eid 2"
        )
        self.simplify(root)
        stmt = root.children[0]
        self.assertEqual(stmt.defined_vars["U"].valuable_references(), 3)
        copy = stmts.Select()
        copy.append_selected(stmt.selection[0].copy(copy))
        copy.append_selected(stmt.selection[1].copy(copy))
        copy.set_where(stmt.where.copy(copy))
        newroot = stmts.Union()
        newroot.append(copy)
        self.annotate(newroot)
        self.simplify(newroot)
        self.assertEqual(
            newroot.as_string(),
            "Any 1,U WHERE 2 owned_by U, NOT EXISTS(1 owned_by U)",
        )
        self.assertEqual(copy.defined_vars["U"].valuable_references(), 3)


class AnnotateTest(unittest.TestCase):
    def setUp(self):
        helper = RQLHelper(DummySchema(), None, {"eid": "uid"})
        self.parse = helper.parse

    def test_is_rel_no_scope_1(self):
        """is relation used as type restriction should not affect variable's
        scope, and should not be included in stinfo['relations']
        """
        rqlst = self.parse("Any X WHERE C is Company, EXISTS(X work_for C)").children[0]
        C = rqlst.defined_vars["C"]
        self.assertIsNot(C.scope, rqlst, C.scope)
        self.assertEqual(len(C.stinfo["relations"]), 1)

    def test_is_rel_no_scope_2(self):
        rqlst = self.parse("Any X, ET WHERE C is ET, EXISTS(X work_for C)").children[0]
        C = rqlst.defined_vars["C"]
        self.assertIs(C.scope, rqlst, C.scope)
        self.assertEqual(len(C.stinfo["relations"]), 2)

    def test_not_rel_normalization_1(self):
        rqlst = self.parse("Any X WHERE C is Company, NOT X work_for C").children[0]
        self.assertEqual(
            rqlst.as_string(), "Any X WHERE C is Company, NOT EXISTS(X work_for C)"
        )
        C = rqlst.defined_vars["C"]
        self.assertIsNot(C.scope, rqlst, C.scope)

    def test_not_rel_normalization_2(self):
        rqlst = self.parse("Any X, ET WHERE C is ET, NOT X work_for C").children[0]
        self.assertEqual(
            rqlst.as_string(), "Any X,ET WHERE C is ET, NOT EXISTS(X work_for C)"
        )
        C = rqlst.defined_vars["C"]
        self.assertIs(C.scope, rqlst, C.scope)

    def test_not_rel_normalization_3(self):
        rqlst = self.parse(
            'Any X WHERE C is Company, X work_for C, NOT C name "World Company"'
        ).children[0]
        self.assertEqual(
            rqlst.as_string(),
            'Any X WHERE C is Company, X work_for C, NOT C name "World Company"',
        )
        C = rqlst.defined_vars["C"]
        self.assertIs(C.scope, rqlst, C.scope)

    def test_not_rel_normalization_4(self):
        rqlst = self.parse(
            'Any X WHERE C is Company, NOT (X work_for C, C name "World Company")'
        ).children[0]
        self.assertEqual(
            rqlst.as_string(),
            'Any X WHERE C is Company, NOT EXISTS(X work_for C, C name "World Company")',
        )
        C = rqlst.defined_vars["C"]
        self.assertIsNot(C.scope, rqlst, C.scope)

    def test_not_rel_normalization_5(self):
        rqlst = self.parse(
            "Any X WHERE X work_for C, "
            'EXISTS(C identity D, NOT Y work_for D, D name "World Company")'
        ).children[0]
        self.assertEqual(
            rqlst.as_string(),
            "Any X WHERE X work_for C, "
            'EXISTS(C identity D, NOT EXISTS(Y work_for D), D name "World Company")',
        )
        D = rqlst.defined_vars["D"]
        D = rqlst.defined_vars["D"]
        self.assertIsNot(D.scope, rqlst, D.scope)
        self.assertIs(D.scope.parent.scope, rqlst, D.scope.parent.scope)

    def test_subquery_annotation_1(self):
        rqlst = self.parse(
            "Any X WITH X BEING (Any X WHERE C is Company, EXISTS(X work_for C))"
        ).children[0]
        C = rqlst.with_[0].query.children[0].defined_vars["C"]
        self.assertIsNot(C.scope, rqlst, C.scope)
        self.assertEqual(len(C.stinfo["relations"]), 1)

    def test_subquery_annotation_2(self):
        rqlst = self.parse(
            "Any X,ET WITH X,ET BEING (Any X, ET WHERE C is ET, EXISTS(X work_for C))"
        ).children[0]
        C = rqlst.with_[0].query.children[0].defined_vars["C"]
        self.assertIs(C.scope, rqlst.with_[0].query.children[0], C.scope)
        self.assertEqual(len(C.stinfo["relations"]), 2)
        X = rqlst.get_variable("X")
        self.assertIs(X.scope, rqlst, X.scope)

    def test_no_attr_var_if_uid_rel(self):
        with self.assertRaises(BadRQLQuery) as cm:
            self.parse("Any X, Y WHERE X work_for Z, Y work_for Z, X eid > Y")
        self.assertEqual(
            str(cm.exception),
            "variable Y should not be used as rhs of attribute relation X eid > Y",
        )

    def test_no_uid_rel_if_not_constant(self):
        rqlst = self.parse("Any X,EID WHERE X eid EID").children[0]
        self.assertEqual(rqlst.defined_vars["X"].stinfo["uidrel"], None)
        rqlst = self.parse("Any X WHERE X eid IN (1,2,3)").children[0]
        self.assertTrue(rqlst.defined_vars["X"].stinfo["uidrel"])

    def test_has_aggregat_false_when_no_aggregat(self):
        rqlst = self.parse("Any ABS(X) WHERE X is Person").children[0]
        self.assertFalse(rqlst.has_aggregat)

    def test_nested_aggregat_detected(self):
        rqlst = self.parse("Any ABS(COUNT(X)) WHERE X is Person").children[0]
        self.assertTrue(rqlst.has_aggregat)


if __name__ == "__main__":
    unittest.main()



# copyright 2004-2022 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of rql.
#
# rql is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# rql is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with rql. If not, see <https://www.gnu.org/licenses/>.
"""RQL Syntax tree annotator"""

from logilab.common.graph import has_path
from logilab.database import UnknownFunction

from rql._exceptions import BadRQLQuery
from rql.utils import function_description
from rql.nodes import (
    Relation,
    VariableRef,
    Constant,
    Not,
    Exists,
    Function,
    And,
    Comparison,
    variable_refs,
    make_relation,
)
from rql.base import BaseNode

from typing import (
    TYPE_CHECKING,
    Dict,
    List,
    Union as Union_,
    Any,
    Optional,
    cast,
)

__docformat__: str = "restructuredtext en"

if TYPE_CHECKING:
    # Cannot find implementation or library stub for module named "unittest_analyze"
    from unittest_analyze import DummySchema  # type:ignore[import]
    import rql


def _var_graphid(
    subvarname: str, trmap: Dict[str, str], select: "rql.stmts.Select"
) -> str:
    try:
        return trmap[subvarname]
    except KeyError:
        return subvarname + str(id(select))


def bloc_simplification(
    variable: Union_["rql.nodes.ColumnAlias", "rql.nodes.Variable"],
    term: "rql.nodes.Relation",
) -> None:
    try:
        variable.stinfo["blocsimplification"].add(term)
    except KeyError:
        variable.stinfo["blocsimplification"] = set((term,))


class GoTo(Exception):
    """Exception used to control the visit of the tree."""

    def __init__(self, node: "rql.nodes.Relation") -> None:
        self.node = node


VAR_SELECTED: int = 1
VAR_HAS_TYPE_REL: int = 2
VAR_HAS_UID_REL: int = 4
VAR_HAS_REL: int = 8


class STCheckState:
    def __init__(self):
        self.errors: List[str] = []
        self.under_not: List[bool] = []
        self.var_info: Dict[
            Union_["rql.nodes.ColumnAlias", "rql.nodes.Variable"], int
        ] = {}

    def error(self, msg: str) -> None:
        self.errors.append(msg)

    def add_var_info(
        self, var: Union_["rql.nodes.ColumnAlias", "rql.nodes.Variable"], vi: int
    ) -> None:
        try:
            self.var_info[var] |= vi
        except KeyError:
            self.var_info[var] = vi


class RQLSTChecker:
    """Check a RQL syntax tree for errors not detected on parsing.

    Some simple rewriting of the tree may be done too:
    * if a OR is used on a symmetric relation
    * IN function with a single child

    use assertions for internal error but specific `BadRQLQuery` exception for
    errors due to a bad rql input
    """

    def __init__(
        self,
        schema: "DummySchema",  # in practice this is a yams Schema
        special_relations: Optional[Dict[str, str]] = None,
        backend: Optional[Any] = None,
    ) -> None:
        self.schema = schema
        self.special_relations = special_relations or {}
        self.backend = backend

    def check(self, node: BaseNode) -> None:
        state = STCheckState()
        self._visit(node, state)
        if state.errors:
            raise BadRQLQuery("%s\n** %s" % (node, "\n** ".join(state.errors)))
        # if node.TYPE == 'select' and \
        #       not node.defined_vars and not node.get_restriction():
        #    result = []
        #    for term in node.selected_terms():
        #        result.append(term.eval(kwargs))

    def _visit(self, node: BaseNode, state: Optional["STCheckState"]) -> None:
        try:
            node.accept(self, state)
        except GoTo as ex:
            self._visit(ex.node, state)
        else:
            for c in node.children:
                self._visit(c, state)
            node.leave(self, state)

    def _visit_selectedterm(
        self,
        node: Union_[
            "rql.stmts.Set", "rql.stmts.Insert", "rql.stmts.Delete", "rql.stmts.Select"
        ],
        state: "STCheckState",
    ) -> None:
        for i, term in enumerate(node.selection):
            # selected terms are not included by the default visit,
            # accept manually each of them
            self._visit(term, state)

    def _check_selected(
        self,
        term: "rql.nodes.VariableRef",
        termtype: str,
        state: "STCheckState",
    ) -> None:
        """check that variables referenced in the given term are selected"""
        for variable_ref in variable_refs(term):
            # no stinfo yet, use references
            # Item "BaseNode" of "Union[BaseNode, VariableRef]" has no attribute "variable"
            variable_ref = cast("rql.nodes.VariableRef", variable_ref)
            for other_variable_ref in variable_ref.variable.references():
                relation: Optional["rql.nodes.Relation"] = other_variable_ref.relation()
                if relation is not None:
                    break
            else:
                message: str = (
                    "variable %s used in %s is not referenced by any relation"
                )
                state.error(message % (variable_ref.name, termtype))

    # statement nodes #########################################################

    def visit_union(self, node: "rql.stmts.Union", state: "STCheckState") -> None:
        # "BaseNode" has no attribute "selection"  [attr-defined]
        nbselected: int = len(cast("rql.stmts.Select", node.children[0]).selection)
        for select in node.children[1:]:
            select = cast("rql.stmts.Select", select)
            if not len(select.selection) == nbselected:
                state.error(
                    "when using union, all subqueries should have "
                    "the same number of selected terms"
                )

    def leave_union(self, node, state):
        pass

    def visit_select(self, node: "rql.stmts.Select", state: "STCheckState") -> None:
        node.vargraph = {}  # graph representing links between variable
        # "Select" has no attribute "aggregated"
        node.aggregated = set()  # type:ignore[attr-defined]
        self._visit_selectedterm(node, state)

    def leave_select(self, node: "rql.stmts.Select", state: "STCheckState") -> None:
        selected = node.selection
        # check selected variable are used in restriction
        if node.where is not None or len(selected) > 1:
            for term in selected:
                self._check_selected(term, "selection", state)
                for vref in term.iget_nodes(VariableRef):
                    state.add_var_info(vref.variable, VAR_SELECTED)
        for var in node.defined_vars.values():
            vinfo: int = state.var_info.get(var, 0)
            if (
                not (vinfo & VAR_HAS_REL)
                and (vinfo & VAR_HAS_TYPE_REL)
                and not (vinfo & VAR_SELECTED)
            ):
                raise BadRQLQuery(f"unbound variable {var.name} ({selected})")
        if node.groupby:
            # check that selected variables are used in groups
            for var in node.selection:
                if isinstance(var, VariableRef) and not any(
                    var.is_equivalent(g) for g in node.groupby
                ):
                    state.error(f"variable {var} should be grouped")
            for group in node.groupby:
                self._check_selected(group, "group", state)
        if node.distinct and node.orderby:
            # check that variables referenced in the given term are reachable from
            # a selected variable with only ?1 cardinality selected
            selectidx = frozenset(
                vref.name for term in selected for vref in term.get_nodes(VariableRef)
            )
            for sortterm in node.orderby:
                for vref in sortterm.term.get_nodes(VariableRef):
                    if vref.name in selectidx:
                        continue
                    for vname in selectidx:
                        try:
                            if self.has_unique_value_path(node, vname, vref.name):
                                break
                        except KeyError:
                            continue  # unlinked variable (usually from a subquery)
                    else:
                        msg = (
                            "can't sort on variable %s which is linked to a"
                            " variable in the selection but may have different"
                            " values for a resulting row"
                        )
                        state.error(msg % vref.name)

    def has_unique_value_path(
        self, select: "rql.stmts.Select", fromvar: str, tovar: str
    ) -> bool:
        graph = select.vargraph
        path: Optional[List[str]] = has_path(
            graph, fromvar, tovar  # type:ignore[arg-type]
        )
        if path is None:
            return False
        for var in path:
            try:
                rtype = graph[(fromvar, var)]
                cardidx: int = 0
            except KeyError:
                rtype = graph[(var, fromvar)]
                cardidx = 1
            rschema: Any = self.schema.rschema(rtype)
            for rdef in rschema.rdefs.values():
                # XXX aggregats handling needs much probably some enhancements...
                # "Select" has no attribute "aggregated"
                if not (
                    var in select.aggregated  # type:ignore[attr-defined]
                    or (
                        rdef.cardinality[cardidx] in "?1"
                        and (var == tovar or not rschema.final)
                    )
                ):
                    return False
            fromvar = var
        return True

    def visit_insert(
        self,
        insert: Union_[
            "rql.stmts.Set", "rql.stmts.Insert", "rql.stmts.Delete", "rql.stmts.Select"
        ],
        state: "STCheckState",
    ) -> None:
        self._visit_selectedterm(insert, state)

    def leave_insert(self, node, state):
        pass

    def visit_delete(self, delete, state):
        self._visit_selectedterm(delete, state)

    def leave_delete(self, node, state):
        pass

    def visit_set(
        self,
        update: Union_[
            "rql.stmts.Set", "rql.stmts.Insert", "rql.stmts.Delete", "rql.stmts.Select"
        ],
        state: "STCheckState",
    ) -> None:
        self._visit_selectedterm(update, state)

    def leave_set(self, node, state):
        pass

    # tree nodes ##############################################################

    def visit_exists(self, node, state):
        pass

    def leave_exists(self, node, state):
        pass

    def visit_subquery(self, node, state):
        pass

    def leave_subquery(self, node, state):
        # copy graph information we're interested in
        pgraph = node.parent.vargraph
        for select in node.query.children:
            # map subquery variable names to outer query variable names
            trmap = {}
            for i, vref in enumerate(node.aliases):
                try:
                    subvref = select.selection[i]
                except IndexError:
                    state.error(
                        'subquery "%s" has only %s selected terms, needs %s'
                        % (select, len(select.selection), len(node.aliases))
                    )
                    continue
                if isinstance(subvref, VariableRef):
                    trmap[subvref.name] = vref.name
                elif (
                    isinstance(subvref, Function)
                    and subvref.descr().aggregat
                    and len(subvref.children) == 1
                    and isinstance(subvref.children[0], VariableRef)
                ):
                    # XXX ok for MIN, MAX, but what about COUNT, AVG...
                    trmap[subvref.children[0].name] = vref.name
                    node.parent.aggregated.add(vref.name)
            for key, val in select.vargraph.items():
                if isinstance(key, tuple):
                    key = (
                        _var_graphid(key[0], trmap, select),
                        _var_graphid(key[1], trmap, select),
                    )
                    pgraph[key] = val
                else:
                    values = pgraph.setdefault(_var_graphid(key, trmap, select), [])
                    values += [_var_graphid(v, trmap, select) for v in val]

    def visit_sortterm(self, sortterm, state):
        term = sortterm.term
        if isinstance(term, Constant):
            for select in sortterm.root.children:
                if len(select.selection) < term.value:
                    state.error(f"order column out of bound {term.value}")
        else:
            stmt = term.stmt
            for tvref in variable_refs(term):
                for vref in tvref.variable.references():
                    if vref.relation() or any(
                        vref.is_equivalent(s) for s in stmt.selection
                    ):
                        break
                else:
                    msg = "sort variable %s is not referenced any where else"
                    state.error(msg % tvref.name)

    def leave_sortterm(self, node, state):
        pass

    def visit_and(self, et, state):
        pass  # assert len(et.children) == 2, len(et.children)

    def leave_and(self, node, state):
        pass

    def visit_or(self, ou, state):
        # assert len(ou.children) == 2, len(ou.children)
        # simplify Ored expression of a symmetric relation
        r1, r2 = ou.children[0], ou.children[1]
        try:
            r1type = r1.r_type
            r2type = r2.r_type
        except AttributeError:
            return  # can't be
        if r1type == r2type and self.schema.relation_schema_for(r1type).symmetric:
            lhs1, rhs1 = r1.get_variable_parts()
            lhs2, rhs2 = r2.get_variable_parts()
            try:
                if lhs1.variable is rhs2.variable and rhs1.variable is lhs2.variable:
                    ou.parent.replace(ou, r1)
                    for vref in r2.get_nodes(VariableRef):
                        vref.unregister_reference()
                    raise GoTo(r1)
            except AttributeError:
                pass

    def leave_or(self, node, state):
        pass

    def visit_not(self, not_, state):
        state.under_not.append(True)

    def leave_not(self, not_, state):
        state.under_not.pop()
        # NOT normalization
        child = not_.children[0]
        if self._should_wrap_by_exists(child):
            not_.replace(child, Exists(child))

    def _should_wrap_by_exists(self, child):
        if isinstance(child, Exists):
            return False
        if not isinstance(child, Relation):
            return True
        if child.r_type == "identity":
            return False
        rschema = self.schema.relation_schema_for(child.r_type)
        if rschema.final:
            return False
        # XXX no exists for `inlined` relation (allow IS NULL optimization)
        # unless the lhs variable is only referenced from this neged relation,
        # in which case it's *not* in the statement's scope, hence EXISTS should
        # be added anyway
        if rschema.inlined:
            references = child.children[0].variable.references()
            valuable = 0
            for vref in references:
                rel = vref.relation()
                if rel is None or not rel.is_types_restriction():
                    if valuable:
                        return False
                    valuable = 1
            return True
        return not child.is_types_restriction()

    def visit_relation(self, relation, state):
        if relation.optional and state.under_not:
            state.error(
                f"can't use optional relation under NOT ({relation.as_string()})"
            )
        lhsvar = relation.children[0].variable
        if relation.is_types_restriction():
            if relation.optional:
                state.error(f'can\'t use optional relation on "{relation.as_string()}"')
            if state.var_info.get(lhsvar, 0) & VAR_HAS_TYPE_REL:
                state.error(
                    "can only one type restriction per variable (use "
                    "IN for %s if desired)" % lhsvar.name
                )
            else:
                state.add_var_info(lhsvar, VAR_HAS_TYPE_REL)
            # special case "C is NULL"
            # if relation.children[1].operator == 'IS':
            #     lhs, rhs = relation.children
            #     #assert isinstance(lhs, VariableRef), lhs
            #     #assert isinstance(rhs.children[0], Constant)
            #     #assert rhs.operator == 'IS', rhs.operator
            #     #assert rhs.children[0].type == None
        else:
            state.add_var_info(lhsvar, VAR_HAS_REL)
            rtype = relation.r_type
            try:
                rschema = self.schema.relation_schema_for(rtype)
            except KeyError:
                state.error(f"unknown relation `{rtype}`")
            else:
                if rschema.final and relation.optional not in (None, "right"):
                    state.error(
                        "optional may only be set on the rhs on final relation `%s`"
                        % relation.r_type
                    )
                if (
                    self.special_relations.get(rtype) == "uid"
                    and relation.operator() == "="
                ):
                    if state.var_info.get(lhsvar, 0) & VAR_HAS_UID_REL:
                        state.error(
                            "can only one uid restriction per variable "
                            "(use IN for %s if desired)" % lhsvar.name
                        )
                    else:
                        state.add_var_info(lhsvar, VAR_HAS_UID_REL)

            for vref in relation.children[1].get_nodes(VariableRef):
                state.add_var_info(vref.variable, VAR_HAS_REL)
        try:
            vargraph = relation.stmt.vargraph
            rhsvarname = relation.children[1].children[0].variable.name
        except AttributeError:
            pass
        else:
            vargraph.setdefault(lhsvar.name, []).append(rhsvarname)
            vargraph.setdefault(rhsvarname, []).append(lhsvar.name)
            vargraph[(lhsvar.name, rhsvarname)] = relation.r_type

    def leave_relation(self, relation, state):
        pass
        # assert isinstance(lhs, VariableRef), '%s: %s' % (lhs.__class__,
        #                                                       relation)

    def visit_comparison(self, comparison, state):
        pass  # assert len(comparison.children) in (1,2), len(comparison.children)

    def leave_comparison(self, node, state):
        pass

    def visit_mathexpression(self, mathexpr, state):
        pass  # assert len(mathexpr.children) == 2, len(mathexpr.children)

    def leave_mathexpression(self, node, state):
        pass

    def visit_unaryexpression(self, unaryexpr, state):
        pass  # assert len(unaryexpr.children) == 2, len(unaryexpr.children)

    def leave_unaryexpression(self, node, state):
        pass

    def visit_function(self, function, state):
        try:
            funcdescr = function_description(function.name)
        except UnknownFunction:
            state.error(f'unknown function "{function.name}"')
        else:
            try:
                funcdescr.check_nbargs(len(function.children))
            except BadRQLQuery as ex:
                state.error(str(ex))
            if self.backend is not None:
                try:
                    funcdescr.st_check_backend(self.backend, function)
                except BadRQLQuery as ex:
                    state.error(str(ex))
            if funcdescr.aggregat:
                if (
                    isinstance(function.children[0], Function)
                    and function.children[0].descr().aggregat
                ):
                    state.error("can't nest aggregat functions")
            if funcdescr.name == "IN":
                # assert function.parent.operator == '='
                if len(function.children) == 1:
                    function.parent.append(function.children[0])
                    function.parent.remove(function)
                # else:
                #    assert len(function.children) >= 1

    def leave_function(self, node, state):
        pass

    def visit_variableref(self, variableref, state):
        # assert len(variableref.children)==0
        # assert not variableref.parent is variableref
        pass

    def visit_variablerefattributeaccess(self, variableref, state):
        pass

    def visit_variablerefmethodcall(self, variableref, state):
        pass

    def leave_variableref(self, node, state):
        pass

    def leave_variablerefattributeaccess(self, node, state):
        pass

    def leave_variablerefmethodcall(self, node, state):
        pass

    def visit_constant(self, constant, state):
        if constant.type != "etype":
            return
        if constant.value not in self.schema:
            state.error(f"unknown entity type {constant.value}")
        if isinstance(constant.parent, Function) and constant.parent.name == "CAST":
            return
        rel = constant.relation()
        if rel is not None and rel.r_type in ("is", "is_instance_of"):
            return
        state.error(
            "Entity types can only be used inside a CAST() " 'or with "is" relation'
        )

    def leave_constant(self, node, state):
        pass


def _check_aggregat_rec(node: "rql.base.BaseNode") -> bool:
    for func in node.iget_nodes(Function):
        if func.descr().aggregat:
            return True
    for child in node.children:
        has_aggregat = _check_aggregat_rec(child)
        if has_aggregat:
            return True
    return False


class RQLSTAnnotator:
    """Annotate RQL syntax tree to ease further code generation from it.

    If an optional variable is shared among multiple scopes, it's rewritten to
    use identity relation.
    """

    def __init__(self, schema, special_relations=None):
        self.schema = schema
        self.special_relations = special_relations or {}

    def annotate(self, node):
        # assert not node.annotated
        node.accept(self)
        node.annotated = True

    def _visit_stmt(self, node):
        for var in node.defined_vars.values():
            var.prepare_annotation()
        for i, term in enumerate(node.selection):
            if _check_aggregat_rec(term):
                node.has_aggregat = True
            # register the selection column index
            for vref in term.get_nodes(VariableRef):
                vref.variable.stinfo["selected"].add(i)
                vref.variable.set_scope(node)
        if node.where is not None:
            node.where.accept(self, node)

    visit_insert = visit_delete = visit_set = _visit_stmt

    def visit_union(self, node):
        for select in node.children:
            self.visit_select(select)

    def visit_select(self, node):
        for var in node.aliases.values():
            var.prepare_annotation()
        if node.with_ is not None:
            for subquery in node.with_:
                self.visit_union(subquery.query)
                subquery.query.schema = node.root.schema
        node.has_aggregat = False
        self._visit_stmt(node)
        if node.having:
            # if there is a having clause, bloc simplification of variables used in GROUPBY
            for term in node.groupby:
                for vref in term.get_nodes(VariableRef):
                    bloc_simplification(vref.variable, term)
            try:
                vargraph = node.vargraph
            except AttributeError:
                vargraph = None
            # XXX node.having is a list of size 1
            assert len(node.having) == 1
            for term in node.having[0].get_nodes(Comparison):
                lhsvariables = set(
                    vref.variable for vref in term.children[0].get_nodes(VariableRef)
                )
                rhsvariables = set(
                    vref.variable for vref in term.children[1].get_nodes(VariableRef)
                )
                for var in lhsvariables | rhsvariables:
                    var.stinfo.setdefault("having", []).append(term)
                if vargraph is not None:
                    for v1 in lhsvariables:
                        v1 = v1.name
                        for v2 in rhsvariables:
                            v2 = v2.name
                            if v1 != v2:
                                vargraph.setdefault(v1, []).append(v2)
                                vargraph.setdefault(v2, []).append(v1)
                if term.optional in ("left", "both"):
                    for var in lhsvariables:
                        if var.stinfo["attrvar"] is not None:
                            optcomps = var.stinfo["attrvar"].stinfo.setdefault(
                                "optcomparisons", set()
                            )
                            optcomps.add(term)
                if term.optional in ("right", "both"):
                    for var in rhsvariables:
                        if var.stinfo["attrvar"] is not None:
                            optcomps = var.stinfo["attrvar"].stinfo.setdefault(
                                "optcomparisons", set()
                            )
                            optcomps.add(term)

    def rewrite_shared_optional(self, exists, var, identity_rel_scope=None):
        """if variable is shared across multiple scopes, need some tree
        rewriting
        """
        # allocate a new variable
        newvar = var.stmt.make_variable()
        newvar.prepare_annotation()
        for vref in var.references():
            if vref.scope is exists:
                rel = vref.relation()
                vref.unregister_reference()
                newvref = VariableRef(newvar)
                vref.parent.replace(vref, newvref)
                stinfo = var.stinfo
                # update stinfo structure which may have already been
                # partially processed
                if rel in stinfo["rhsrelations"]:
                    lhs, rhs = rel.get_parts()
                    if (
                        vref is rhs.children[0]
                        and self.schema.rschema(rel.r_type).final
                    ):
                        update_attrvars(newvar, rel, lhs)
                        lhsvar = getattr(lhs, "variable", None)
                        stinfo["attrvars"].remove((lhsvar, rel.r_type))
                        if stinfo["attrvar"] is lhsvar:
                            if stinfo["attrvars"]:
                                stinfo["attrvar"] = next(iter(stinfo["attrvars"]))
                            else:
                                stinfo["attrvar"] = None
                    stinfo["rhsrelations"].remove(rel)
                    newvar.stinfo["rhsrelations"].add(rel)
                try:
                    stinfo["relations"].remove(rel)
                    newvar.stinfo["relations"].add(rel)
                except KeyError:
                    pass
                try:
                    stinfo["optrelations"].remove(rel)
                    newvar.add_optional_relation(rel)
                except KeyError:
                    pass
                try:
                    stinfo["blocsimplification"].remove(rel)
                    bloc_simplification(newvar, rel)
                except KeyError:
                    pass
                if stinfo["uidrel"] is rel:
                    newvar.stinfo["uidrel"] = rel
                    stinfo["uidrel"] = None
                if stinfo["typerel"] is rel:
                    newvar.stinfo["typerel"] = rel
                    stinfo["typerel"] = None
        # shared references
        newvar.stinfo["constnode"] = var.stinfo["constnode"]
        if newvar.stmt.solutions:  # solutions already computed
            newvar.stinfo["possibletypes"] = var.stinfo["possibletypes"]
            for sol in newvar.stmt.solutions:
                sol[newvar.name] = sol[var.name]
        if identity_rel_scope is None:
            rel = exists.add_relation(var, "identity", newvar)
            identity_rel_scope = exists
        else:
            rel = make_relation(var, "identity", (newvar,), VariableRef)
            exists.parent.replace(exists, And(exists, Exists(rel)))
        # we have to force visit of the introduced relation
        self.visit_relation(rel, identity_rel_scope)
        return newvar

    # tree nodes ##############################################################

    def visit_exists(self, node, scope):
        node.children[0].accept(self, node)

    def visit_not(self, node, scope):
        node.children[0].accept(self, scope)

    def visit_and(self, node, scope):
        node.children[0].accept(self, scope)
        node.children[1].accept(self, scope)

    visit_or = visit_and

    def visit_relation(self, relation, scope):
        # assert relation.parent, repr(relation)
        lhs, rhs = relation.get_parts()
        # may be a constant once rqlst has been simplified
        lhsvar = getattr(lhs, "variable", None)
        if relation.is_types_restriction():
            if lhsvar is not None:
                lhsvar.stinfo["typerel"] = relation
            return
        if relation.optional is not None:
            exists = relation.scope
            if not isinstance(exists, Exists):
                exists = None
            if lhsvar is not None:
                if exists is not None and lhsvar.scope is lhsvar.stmt:
                    lhsvar = self.rewrite_shared_optional(exists, lhsvar)
                bloc_simplification(lhsvar, relation)
                if relation.optional == "both":
                    lhsvar.add_optional_relation(relation)
                elif relation.optional == "left":
                    lhsvar.add_optional_relation(relation)
            try:
                rhsvar = rhs.children[0].variable
                if exists is not None and rhsvar.scope is rhsvar.stmt:
                    rhsvar = self.rewrite_shared_optional(exists, rhsvar)
                bloc_simplification(rhsvar, relation)
                if relation.optional == "right":
                    rhsvar.add_optional_relation(relation)
                elif relation.optional == "both":
                    rhsvar.add_optional_relation(relation)
            except AttributeError:
                # may have been rewritten as well
                pass
        rtype = relation.r_type
        rschema = self.schema.relation_schema_for(rtype)
        if lhsvar is not None:
            lhsvar.set_scope(scope)
            lhsvar.stinfo["relations"].add(relation)
            if rtype in self.special_relations:
                key = f"{self.special_relations[rtype]}rels"
                if key == "uidrels":
                    constnode = relation.get_variable_parts()[1]
                    if not (
                        relation.operator() != "="
                        # XXX use state to detect relation under NOT/OR
                        # + check variable's scope
                        or isinstance(relation.parent, Not)
                        or relation.parent.ored()
                        # NOT EXISTS should be treated like NOT
                        or (
                            isinstance(relation.parent, Exists)
                            and isinstance(relation.parent.parent, Not)
                        )
                    ):
                        if isinstance(constnode, Constant):
                            lhsvar.stinfo["constnode"] = constnode
                        if not isinstance(constnode, VariableRef):
                            lhsvar.stinfo["uidrel"] = relation
                else:
                    lhsvar.stinfo.setdefault(key, set()).add(relation)
            elif rschema.final or rschema.inlined:
                bloc_simplification(lhsvar, relation)
        for vref in rhs.get_nodes(VariableRef):
            var = vref.variable
            var.set_scope(scope)
            var.stinfo["relations"].add(relation)
            var.stinfo["rhsrelations"].add(relation)
            if vref is rhs.children[0] and rschema.final:
                update_attrvars(var, relation, lhs)


def update_attrvars(var, relation, lhs):
    if var.stinfo["relations"] - var.stinfo["rhsrelations"]:
        raise BadRQLQuery(
            "variable %s should not be used as rhs of attribute relation %s"
            % (var.name, relation)
        )
    # stinfo['attrvars'] is set of couple (lhs variable name, relation name)
    # where the `var` attribute variable is used
    lhsvar = getattr(lhs, "variable", None)
    try:
        var.stinfo["attrvars"].add((lhsvar, relation.r_type))
    except KeyError:
        var.stinfo["attrvars"] = set([(lhsvar, relation.r_type)])
    # give priority to variable which is not in an EXISTS as
    # "main" attribute variable
    if var.stinfo["attrvar"] is None or not isinstance(relation.scope, Exists):
        var.stinfo["attrvar"] = lhsvar or lhs
